import logging
from collections import defaultdict
from typing import Any, Literal

from pydantic import Field, PrivateAttr, ValidationInfo, model_validator

from hexdoc.core import (
    ItemStack,
    ModResourceLoader,
    ResLoc,
    ResourceLocation,
)
from hexdoc.core.compat import AtLeast_1_20, Before_1_20
from hexdoc.minecraft import LocalizedStr
from hexdoc.model import Color, HexdocModel
from hexdoc.utils import ContextSource, cast_context, sorted_dict

from .book_context import BookContext
from .category import Category
from .entry import Entry
from .text import FormattingContext, FormatTree

logger = logging.getLogger(__name__)


class Book(HexdocModel):
    """Main Patchouli book class.

    Includes all data from book.json, categories/entries/pages, and i18n.

    You should probably not use this (or any other Patchouli types, eg. Category, Entry)
    to edit and re-serialize book.json, because this class sets all the default values
    as defined by the docs.

    See: https://vazkiimods.github.io/Patchouli/docs/reference/book-json
    """

    # not in book.json
    _categories: dict[ResourceLocation, Category] = PrivateAttr(default_factory=dict)
    _all_entries: dict[ResourceLocation, Entry] = PrivateAttr(default_factory=dict)

    # required
    name: LocalizedStr
    landing_text: FormatTree

    # required in 1.20 but optional in 1.19
    # so we'll make it optional and validate later
    use_resource_pack: AtLeast_1_20[Literal[True]] | Before_1_20[bool] = False

    # optional
    book_texture: ResourceLocation = ResLoc("patchouli", "textures/gui/book_brown.png")
    filler_texture: ResourceLocation | None = None
    crafting_texture: ResourceLocation | None = None
    model: ResourceLocation = ResLoc("patchouli", "book_brown")
    text_color: Color = Color("000000")
    header_color: Color = Color("333333")
    nameplate_color: Color = Color("FFDD00")
    link_color: Color = Color("0000EE")
    link_hover_color: Color = Color("8800EE")
    progress_bar_color: Color = Color("FFFF55")
    progress_bar_background: Color = Color("DDDDDD")
    open_sound: ResourceLocation | None = None
    flip_sound: ResourceLocation | None = None
    index_icon: ResourceLocation | None = None
    pamphlet: bool = False
    show_progress: bool = True
    version: str | int = 0
    subtitle: LocalizedStr | None = None
    creative_tab: str | None = None
    advancements_tab: str | None = None
    dont_generate_book: bool = False
    custom_book_item: ItemStack | None = None
    show_toasts: bool = True
    use_blocky_font: bool = False
    i18n: bool = False
    macros: dict[str, str] = Field(default_factory=dict)
    pause_game: bool = False
    text_overflow_mode: Literal["overflow", "resize", "truncate"] | None = None

    @property
    def categories(self):
        return self._categories

    @property
    def all_entries(self):
        """
        Note: includes external entries as well (not just internal ones)
        """
        return self._all_entries

    def _load_categories(self, context: ContextSource, book_ctx: BookContext):
        categories = Category.load_all(
            cast_context(context),
            book_ctx.book_id,
            self.use_resource_pack,
        )

        if not categories:
            raise ValueError(
                "No categories found, are the paths in your properties file correct?"
            )

        for id, category in categories.items():
            self._categories[id] = category

            link_base = book_ctx.get_link_base(category.resource_dir)
            book_ctx.book_links[category.book_link_key] = link_base.with_fragment(
                category.fragment
            )

    def _load_entries(
        self,
        context: ContextSource,
        book_ctx: BookContext,
        loader: ModResourceLoader,
    ):
        internal_entries = defaultdict[ResLoc, dict[ResLoc, Entry]](dict)
        spoilered_categories = dict[ResLoc, bool]()

        for resource_dir, id, data in loader.load_book_assets(
            parent_book_id=book_ctx.book_id,
            folder="entries",
            use_resource_pack=self.use_resource_pack,
        ):
            entry = Entry.load(resource_dir, id, data, cast_context(context))
            if not entry.is_flag_enabled:
                logger.info(f"Skipping entry {id} due to disabled flag {entry.flag}")
                continue

            spoilered_categories[entry.category_id] = (
                entry.is_spoiler and spoilered_categories.get(entry.category_id, True)
            )

            # i used the entry to insert the entry (pretty sure thanos said that)
            if resource_dir.internal:
                internal_entries[entry.category_id][entry.id] = entry
            self._all_entries[entry.id] = entry

            link_base = book_ctx.get_link_base(resource_dir)
            book_ctx.book_links[entry.book_link_key] = link_base.with_fragment(
                entry.fragment
            )

            for page in entry.pages:
                page_key = page.book_link_key(entry.book_link_key)
                page_fragment = page.fragment(entry.fragment)
                if page_key is not None and page_fragment is not None:
                    book_ctx.book_links[page_key] = link_base.with_fragment(
                        page_fragment
                    )

        if not internal_entries:
            raise ValueError(
                f"No internal entries found for book {book_ctx.book_id}, is this the correct id?"
            )

        for category_id, new_entries in internal_entries.items():
            category = self._categories[category_id]
            category.entries = sorted_dict(category.entries | new_entries)
            if is_spoiler := spoilered_categories.get(category.id):
                category.is_spoiler = is_spoiler

    @model_validator(mode="before")
    @classmethod
    def _pre_root(cls, data: dict[Any, Any] | Any):
        if isinstance(data, dict) and "index_icon" not in data:
            data["index_icon"] = data.get("model")
        return data

    @model_validator(mode="after")
    def _post_root(self, info: ValidationInfo):
        if not info.context:
            return self

        # make the macros accessible when rendering the template
        self.macros |= FormattingContext.of(info).macros

        return self



"""
Commands designed to run in a GitHub Actions workflow.
"""

# pyright: reportPrivateUsage=false

import os
import shutil
import subprocess
from functools import cached_property
from pathlib import Path
from typing import Literal, TypedDict, TypeVar, Unpack

from github import Auth, Github, UnknownObjectException
from github.Repository import Repository
from pydantic import TypeAdapter
from typer import Typer

from hexdoc.cli.utils.args import PropsOption, ReleaseOption
from hexdoc.model import HexdocModel, HexdocSettings
from hexdoc.utils import setup_logging

app = Typer(name="ci")


@app.command()
def build(
    *,
    props_file: PropsOption,
    release: ReleaseOption,
):
    from . import app as hexdoc_app

    env = CIEnvironment.model_getenv()
    setup_logging(env.verbosity, ci=True)

    # FIXME: scuffed. why are we setting environment variables here :/
    for key in [
        "MOCK_GITHUB_PAGES_URL",  # highest priority so tests work correctly
        "HEXDOC_SITE_URL",  # TODO: this should be supported in more places
        "GITHUB_PAGES_URL",
    ]:
        if pages_url := os.getenv(key):
            break
    else:
        pages_url = get_pages_url(env.repo)
    os.environ["GITHUB_PAGES_URL"] = pages_url

    site_path = hexdoc_app.build(
        Path("_site/src/docs"),
        clean=True,
        branch=env.branch,
        props_file=props_file,
        release=release,
    )

    site_dist = site_path / "dist"
    if site_dist.is_dir():
        shutil.rmtree(site_dist)

    subprocess.run(["hatch", "build", "--clean"], check=True)
    shutil.copytree("dist", site_dist)

    env.set_output("pages-url", pages_url)


@app.command()
def merge(
    *,
    props_file: PropsOption,
    release: ReleaseOption,
):
    from . import app as hexdoc_app

    env = CIEnvironment.model_getenv()
    setup_logging(env.verbosity, ci=True)
    hexdoc_app.merge(props_file=props_file, release=release)


@app.command(deprecated=True)
def export(
    *,
    props_file: PropsOption,
    release: ReleaseOption,
):
    build(props_file=props_file, release=release)


@app.command(deprecated=True)
def render(
    lang: str,
    *,
    props_file: PropsOption,
    release: ReleaseOption,
):
    build(props_file=props_file, release=release)


# utils

_T = TypeVar("_T")


class CIEnvironment(HexdocSettings):
    github_output: str
    github_ref_name: str
    github_repository: str
    github_token: str | None = None
    runner_debug: bool = False

    @property
    def branch(self):
        return self.github_ref_name

    @property
    def verbosity(self):
        return 1 if self.runner_debug else 0

    @cached_property
    def repo(self):
        return self.gh.get_repo(self.github_repository)

    @cached_property
    def gh(self):
        return Github(
            auth=self._gh_auth,
        )

    @property
    def _gh_auth(self):
        if self.github_token:
            return Auth.Token(self.github_token)

    def set_output(self, name: str, value: str | tuple[type[_T], _T]):
        match value:
            case str():
                pass
            case (data_type, data):
                ta = TypeAdapter(data_type)
                value = ta.dump_json(data).decode()

        with open(self.github_output, "a") as f:
            print(f"{name}={value}", file=f)


def get_pages_url(repo: Repository) -> str:
    endpoint = f"{repo.url}/pages"
    try:
        _, data = repo._requester.requestJsonAndCheck("GET", endpoint)
    except Exception as e:
        e.add_note(f"  Endpoint: {endpoint}")
        if isinstance(e, UnknownObjectException):
            e.add_note(
                "Note: check if GitHub Pages is enabled in this repo"
                + " (https://hexdoc.hexxy.media/docs/guides/deployment/github-pages)"
            )
        raise
    return str(data["html_url"])


class CIMatrixItem(HexdocModel):
    value: str
    continue_on_error: bool


class AnnotationKwargs(TypedDict, total=False):
    title: str
    """Custom title"""
    file: str
    """Filename"""
    col: int
    """Column number, starting at 1"""
    endColumn: int
    """End column number"""
    line: int
    """Line number, starting at 1"""
    endLine: int
    """End line number"""


def add_notice(message: str, **kwargs: Unpack[AnnotationKwargs]):
    return add_annotation("notice", message, **kwargs)


def add_warning(message: str, **kwargs: Unpack[AnnotationKwargs]):
    return add_annotation("warning", message, **kwargs)


def add_error(message: str, **kwargs: Unpack[AnnotationKwargs]):
    return add_annotation("error", message, **kwargs)


def add_annotation(
    type: Literal["notice", "warning", "error"],
    message: str,
    **kwargs: Unpack[AnnotationKwargs],
):
    if kwargs:
        kwargs_str = " " + ",".join(f"{k}={v}" for k, v in kwargs.items())
    else:
        kwargs_str = ""

    print(f"::{type}{kwargs_str}::{message}")



from __future__ import annotations

import logging
from collections import defaultdict
from functools import cached_property
from pathlib import Path
from typing import Annotated, Any, Literal, Self, Sequence

from pydantic import Field, PrivateAttr, field_validator, model_validator
from pydantic.json_schema import (
    DEFAULT_REF_TEMPLATE,
    GenerateJsonSchema,
    SkipJsonSchema,
)
from typing_extensions import override
from yarl import URL

from hexdoc.model.base import HexdocSettings
from hexdoc.model.strip_hidden import StripHiddenModel
from hexdoc.utils import (
    TRACE,
    PydanticOrderedSet,
    RelativePath,
    ValidationContext,
    git_root,
    load_toml_with_placeholders,
    relative_path_root,
)
from hexdoc.utils.deserialize.toml import GenerateJsonSchemaTOML
from hexdoc.utils.types import PydanticURL

from .resource import ResourceLocation
from .resource_dir import ResourceDir

logger = logging.getLogger(__name__)

JINJA_NAMESPACE_ALIASES = {
    "patchouli": "hexdoc",
}


class EnvironmentVariableProps(HexdocSettings):
    # default Actions environment variables
    github_repository: str
    github_sha: str

    # set by CI
    github_pages_url: PydanticURL

    # for putting books somewhere other than the site root
    hexdoc_subdirectory: str | None = None

    # optional for debugging
    debug_githubusercontent: PydanticURL | None = None

    @property
    def asset_url(self) -> URL:
        if self.debug_githubusercontent is not None:
            return URL(str(self.debug_githubusercontent))

        return (
            URL("https://raw.githubusercontent.com")
            / self.github_repository
            / self.github_sha
        )

    @property
    def source_url(self) -> URL:
        return (
            URL("https://github.com")
            / self.github_repository
            / "tree"
            / self.github_sha
        )

    @property
    def repo_owner(self):
        return self._github_repository_parts[0]

    @property
    def repo_name(self):
        return self._github_repository_parts[1]

    @property
    def _github_repository_parts(self):
        owner, repo_name = self.github_repository.split("/", maxsplit=1)
        return owner, repo_name

    @model_validator(mode="after")
    def _append_subdirectory(self) -> Self:
        if self.hexdoc_subdirectory:
            self.github_pages_url /= self.hexdoc_subdirectory
        return self


class TemplateProps(StripHiddenModel, validate_assignment=True):
    static_dir: RelativePath | None = None
    icon: RelativePath | None = None
    include: PydanticOrderedSet[str]

    render_from: PydanticOrderedSet[str] = Field(None, validate_default=False)  # type: ignore
    """List of modids to include default rendered templates from.

    If not provided, defaults to `self.include`.
    """
    render: dict[Path, str] = Field(default_factory=dict)
    extend_render: dict[Path, str] = Field(default_factory=dict)

    redirect: tuple[Path, str] | None = (Path("index.html"), "redirect.html.jinja")
    """filename, template"""

    args: dict[str, Any]

    _was_render_set: bool = PrivateAttr(False)

    @property
    def override_default_render(self):
        return self._was_render_set

    @field_validator("include", "render_from", mode="after")
    @classmethod
    def _resolve_aliases(cls, values: PydanticOrderedSet[str] | None):
        if values:
            for alias, replacement in JINJA_NAMESPACE_ALIASES.items():
                if alias in values:
                    values.remove(alias)
                    values.add(replacement)
        return values

    @model_validator(mode="after")
    def _set_default_render_from(self):
        if self.render_from is None:  # pyright: ignore[reportUnnecessaryComparison]
            self.render_from = self.include
        return self


# TODO: support item/block override
class PNGTextureOverride(StripHiddenModel):
    url: PydanticURL
    pixelated: bool


class TextureTextureOverride(StripHiddenModel):
    texture: ResourceLocation
    """The id of an image texture (eg. `minecraft:textures/item/stick.png`)."""


class TexturesProps(StripHiddenModel):
    enabled: bool = True
    """Set to False to disable texture rendering."""
    strict: bool = True
    """Set to False to print some errors instead of throwing them."""
    missing: set[ResourceLocation] | Literal["*"] = Field(default_factory=set)
    override: dict[
        ResourceLocation,
        PNGTextureOverride | TextureTextureOverride,
    ] = Field(default_factory=dict)


class LangProps(StripHiddenModel):
    """Configuration for a specific book language."""

    quiet: bool = False
    """If `True`, do not log warnings for missing translations.

    Using this option for the default language is not recommended.
    """
    ignore_errors: bool = False
    """If `True`, log fatal errors for this language instead of failing entirely.

    Using this option for the default language is not recommended.
    """


class BaseProperties(StripHiddenModel, ValidationContext):
    env: SkipJsonSchema[EnvironmentVariableProps]
    props_dir: SkipJsonSchema[Path]

    @classmethod
    def load(cls, path: Path) -> Self:
        return cls.load_data(
            props_dir=path.parent,
            data=load_toml_with_placeholders(path),
        )

    @classmethod
    def load_data(cls, props_dir: Path, data: dict[str, Any]) -> Self:
        props_dir = props_dir.resolve()

        with relative_path_root(props_dir):
            env = EnvironmentVariableProps.model_getenv()
            props = cls.model_validate(
                data
                | {
                    "env": env,
                    "props_dir": props_dir,
                },
            )

        logger.log(TRACE, props)
        return props

    @override
    @classmethod
    def model_json_schema(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls,
        by_alias: bool = True,
        ref_template: str = DEFAULT_REF_TEMPLATE,
        schema_generator: type[GenerateJsonSchema] = GenerateJsonSchemaTOML,
        mode: Literal["validation", "serialization"] = "validation",
    ) -> dict[str, Any]:
        return super().model_json_schema(by_alias, ref_template, schema_generator, mode)


class Properties(BaseProperties):
    """Pydantic model for `hexdoc.toml` / `properties.toml`."""

    modid: str

    book_type: str = "patchouli"
    """Modid of the `hexdoc.plugin.BookPlugin` to use when loading this book."""

    # TODO: make another properties type without book_id
    book_id: ResourceLocation | None = Field(alias="book", default=None)
    extra_books: list[ResourceLocation] = Field(default_factory=list)

    default_lang: str = "en_us"
    default_branch: str = "main"

    is_0_black: bool = False
    """If true, the style `$(0)` changes the text color to black; otherwise it resets
    the text color to the default."""

    resource_dirs: Sequence[ResourceDir]
    export_dir: RelativePath | None = None

    entry_id_blacklist: set[ResourceLocation] = Field(default_factory=set)

    macros: dict[str, str] = Field(default_factory=dict)
    link_overrides: dict[str, str] = Field(default_factory=dict)

    textures: TexturesProps = Field(default_factory=TexturesProps)

    flags: dict[str, bool] = Field(default_factory=dict)
    """Local Patchouli flag overrides.

    This has the final say over built-in defaults and flags exported by other mods.
    """

    template: TemplateProps | None = None

    lang: defaultdict[
        str,
        Annotated[LangProps, Field(default_factory=LangProps)],
    ] = Field(default_factory=lambda: defaultdict(LangProps))
    """Per-language configuration. The key should be the language code, eg. `en_us`."""

    extra: dict[str, Any] = Field(default_factory=dict)

    def mod_loc(self, path: str) -> ResourceLocation:
        """Returns a ResourceLocation with self.modid as the namespace."""
        return ResourceLocation(self.modid, path)

    @property
    def prerender_dir(self):
        return self.cache_dir / "prerender"

    @property
    def cache_dir(self):
        return self.repo_root / ".hexdoc"

    @cached_property
    def repo_root(self):
        return git_root(self.props_dir)



from pathlib import Path

from hexdoc.utils import commands


def git_root(cwd: str | Path) -> Path:
    return commands.run(
        ["git", "rev-parse", "--show-toplevel"],
        cwd=cwd,
        type=Path,
    )



import contextlib
import inspect
import io
import sys
import traceback
from types import CodeType, FrameType, TracebackType
from typing import cast


class FilteredTracebackWrapper:
    """Wrapper for `types.TracebackType` to allow hiding modules from tracebacks."""

    def __init__(self, tb: TracebackType, hidden_modules: set[str]):
        self._tb = tb
        self._hidden_modules = hidden_modules
        self._skipped = 0

    @classmethod
    def wrap_cast(cls, tb: TracebackType | None, hidden_modules: set[str]):
        if tb is None:
            return None
        return cls(tb, hidden_modules).cast()

    @classmethod
    def _wrap(cls, tb: TracebackType | None, hidden_modules: set[str]):
        if tb is None:
            return None
        return cls(tb, hidden_modules)

    @property
    def tb_frame(self) -> FrameType:
        frame = self._tb.tb_frame
        if not self._skipped:
            return frame
        return SkippedCallsFrameWrapper(frame, self._skipped).cast()

    @property
    def tb_next(self) -> TracebackType | None:
        tb_next = self._wrap_next()
        skipped = 0

        while tb_next:
            if tb_next._is_hidden():
                tb_next = tb_next._wrap_next()
                skipped += 1
                continue

            tb_next._skipped = skipped
            return tb_next.cast()

        return None

    def cast(self):
        return cast(TracebackType, self)

    def _is_hidden(self):
        # never filter out the last frame
        if self._tb.tb_next is None:
            return False

        mod = self._get_module()
        return mod is not None and mod.__name__ in self._hidden_modules

    def _get_module(self):
        return inspect.getmodule(self.tb_frame)

    def _wrap_next(self):
        return self._wrap(self._tb.tb_next, self._hidden_modules)

    def __getattr__(self, name: str):
        return getattr(self._tb, name)


class SkippedCallsFrameWrapper:
    """Wrapper for `types.FrameType` to allow changing read-only attributes."""

    def __init__(self, frame: FrameType, skipped: int):
        self._frame = frame
        self.f_code = SkippedCallsCodeWrapper(frame.f_code, skipped)

    def cast(self):
        return cast(FrameType, self)

    def __getattr__(self, name: str):
        return getattr(self._frame, name)


class SkippedCallsCodeWrapper:
    """Wrapper for `types.CodeType` to allow changing read-only attributes."""

    def __init__(self, code: CodeType, skipped: int):
        self._code = code

        s = "" if skipped == 1 else "s"
        self.co_name = f"{code.co_name} (...after {skipped} skipped call{s})"

    def cast(self):
        return cast(CodeType, self)

    def __getattr__(self, name: str):
        return getattr(self._code, name)


def create_filtered_excepthook(hidden_modules: set[str]):
    """Creates a replacement for `sys.excepthook` using `FilteredTracebackWrapper`.

    Traceback frames where the `__name__` of the source module is in `hidden_modules`
    will not be printed by this hook.
    """

    def filtered_excepthook(
        exc_type: type[BaseException],
        exc_value: BaseException,
        exc_tb: TracebackType | None,
    ):
        exc_tb = FilteredTracebackWrapper.wrap_cast(exc_tb, hidden_modules)
        traceback.print_exception(exc_type, exc_value, exc_tb)

    return filtered_excepthook


# https://github.com/python/cpython/blob/6f9ca53a6ac343a5/Lib/idlelib/run.py#L225
def get_message_with_hints(exc: AttributeError | NameError, removeprefix: bool = True):
    err = io.StringIO()
    with contextlib.redirect_stderr(err):
        sys.__excepthook__(type(exc), exc, None)
    message = err.getvalue().strip().split("\n")[-1]
    if removeprefix:
        message = message.removeprefix(f"{type(exc).__name__}: ")
    return message



"""Common utility functions used throughout Favicons."""

# Standard Library
from pathlib import Path
from tempfile import mkstemp
from typing import Generator, Mapping, Union

from ._constants import ICON_TYPES
from ._exceptions import FaviconNotFoundError, FaviconsError
from ._types import Color, FaviconProperties


def validate_path(
    path: Union[Path, str], must_exist: bool = True, create: bool = False
) -> Path:
    """Validate a path and ensure it's a Path object."""

    if isinstance(path, str):
        try:
            path = Path(path)
        except TypeError as err:
            raise FaviconsError("{path} is not a valid path.", path=path) from err

    if create:
        if path.is_dir() and not path.exists():
            path.mkdir(parents=True)
        elif not path.is_dir() and not path.parent.exists():
            path.parent.mkdir(parents=True)

    if must_exist and not path.exists():
        raise FaviconNotFoundError(path)

    return path


def generate_icon_types() -> Generator[FaviconProperties, None, None]:
    """Get icon type objects."""
    for icon_type in ICON_TYPES:
        if isinstance(icon_type, Mapping):
            yield FaviconProperties(**icon_type)


def svg_to_png(svg_path: Path, background_color: Color) -> Path:
    """Convert an SVG vector to a PNG file."""
    # Third Party
    from reportlab.graphics import renderPM
    from reportlab.lib.colors import transparent
    from svglib.svglib import svg2rlg

    _, png_path = mkstemp(suffix=".tiff")

    png = Path(png_path)

    drawing = svg2rlg(str(svg_path))
    renderPM.drawToFile(
        drawing,
        str(png),
        fmt="TIFF",
        bg=int(background_color.as_hex().replace("#", ""), 16),
        configPIL={"transparent": transparent},
    )

    return png



__all__ = [
    "BlastingPage",
    "CampfireCookingPage",
    "CraftingPage",
    "EmptyPage",
    "EntityPage",
    "ImagePage",
    "LinkPage",
    "Multiblock",
    "MultiblockPage",
    "Page",
    "PageWithText",
    "PageWithTitle",
    "QuestPage",
    "RelationsPage",
    "SmeltingPage",
    "SmithingPage",
    "SmokingPage",
    "SpotlightPage",
    "StonecuttingPage",
    "TextPage",
]

from .abstract_pages import Page, PageWithText, PageWithTitle
from .pages import (
    BlastingPage,
    CampfireCookingPage,
    CraftingPage,
    EmptyPage,
    EntityPage,
    ImagePage,
    LinkPage,
    Multiblock,
    MultiblockPage,
    QuestPage,
    RelationsPage,
    SmeltingPage,
    SmithingPage,
    SmokingPage,
    SpotlightPage,
    StonecuttingPage,
    TextPage,
)



from typing import Any, dataclass_transform

from pydantic import ConfigDict, model_validator
from pydantic.config import JsonDict

from hexdoc.utils.deserialize.assertions import cast_or_raise

from .base import DEFAULT_CONFIG, HexdocModel


def _json_schema_extra(schema: JsonDict):
    cast_or_raise(schema.setdefault("patternProperties", {}), JsonDict).update(
        {
            r"^\_": {},
        },
    )


@dataclass_transform()
class StripHiddenModel(HexdocModel):
    """Base model which removes all keys starting with _ before validation."""

    model_config = DEFAULT_CONFIG | ConfigDict(
        json_schema_extra=_json_schema_extra,
    )

    @model_validator(mode="before")
    def _pre_root_strip_hidden(cls, values: dict[Any, Any] | Any) -> Any:
        if not isinstance(values, dict):
            return values

        return {
            key: value
            for key, value in values.items()
            if not (isinstance(key, str) and (key.startswith("_") or key == "$schema"))
        }



__all__ = [
    "AnimatedTexture",
    "AnimationMeta",
    "HexdocAssetLoader",
    "ImageTexture",
    "ItemTexture",
    "ItemWithTexture",
    "ModelItem",
    "MultiItemTexture",
    "NamedTexture",
    "PNGTexture",
    "SingleItemTexture",
    "TagWithTexture",
    "Texture",
    "TextureContext",
    "TextureLookup",
    "TextureLookups",
    "validate_texture",
]

from .animated import (
    AnimatedTexture,
    AnimationMeta,
)
from .items import (
    ImageTexture,
    ItemTexture,
    MultiItemTexture,
    SingleItemTexture,
)
from .load_assets import (
    HexdocAssetLoader,
    Texture,
    validate_texture,
)
from .models import ModelItem
from .textures import (
    PNGTexture,
    TextureContext,
    TextureLookup,
    TextureLookups,
)
from .with_texture import (
    ItemWithTexture,
    NamedTexture,
    TagWithTexture,
)

HexdocPythonResourceLoader = None
"""PLACEHOLDER - DO NOT USE

This class has been removed from hexdoc, but this variable is required to fix an import
error with old versions of `hexdoc_minecraft`.
"""



from typing import Any, Generic, Iterable, TypeVar

import networkx as nx
from typing_extensions import override

_Node = TypeVar("_Node")


class TypedDiGraph(nx.DiGraph, Generic[_Node]):  # type: ignore
    """A very incomplete wrapper type over `nx.DiGraph` to add some type hints."""

    @override
    def add_edge(self, u_of_edge: _Node, v_of_edge: _Node, **attr: Any) -> None:
        super().add_edge(u_of_edge, v_of_edge, **attr)  # type: ignore

    def iter_out_edges(self, node: _Node) -> Iterable[tuple[_Node, _Node]]:
        return self.out_edges(node)  # type: ignore

    def find_cycle(self) -> list[tuple[_Node, _Node]] | None:
        try:
            return nx.find_cycle(self)  # type: ignore
        except nx.NetworkXNoCycle:
            return None

    def topological_sort(self) -> Iterable[_Node]:
        return nx.topological_sort(self)  # type: ignore



from jinja2 import nodes
from jinja2.ext import Extension
from jinja2.parser import Parser
from markupsafe import Markup


# https://stackoverflow.com/a/64392515
class IncludeRawExtension(Extension):
    tags = {"include_raw"}

    def parse(self, parser: Parser) -> nodes.Node | list[nodes.Node]:
        lineno = parser.stream.expect("name:include_raw").lineno
        template = parser.parse_expression()
        result = self.call_method("_render", [template], lineno=lineno)
        return nodes.Output([result], lineno=lineno)

    def _render(self, filename: str) -> Markup:
        assert self.environment.loader is not None
        source = self.environment.loader.get_source(self.environment, filename)
        return Markup(source[0])


# https://github.com/mitmproxy/pdoc/blob/895dae1895/pdoc/render_helpers.py#L484
class DefaultMacroExtension(Extension):
    """
    This extension provides a new `{% defaultmacro %}` statement, which defines a macro only if it does not exist.

    For example,

    ```html+jinja
    {% defaultmacro example() %}
        test 123
    {% enddefaultmacro %}
    ```

    is equivalent to

    ```html+jinja
    {% macro default_example() %}
    test 123
    {% endmacro %}
    {% if not example %}
        {% macro example() %}
            test 123
        {% endmacro %}
    {% endif %}
    ```

    Additionally, the default implementation is also available as `default_$macroname`, which makes it possible
    to reference it in the override.
    """

    tags = {"defaultmacro"}

    def parse(self, parser: Parser) -> nodes.Node | list[nodes.Node]:
        m = nodes.Macro(lineno=next(parser.stream).lineno)
        name = parser.parse_assign_target(name_only=True).name
        parser.parse_signature(m)
        m.body = parser.parse_statements(("name:enddefaultmacro",), drop_needle=True)

        if_stmt = nodes.If(
            nodes.Not(
                nodes.Test(nodes.Name(name, "load"), "defined", [], [], None, None)
            ),
            [nodes.Macro(name, m.args, m.defaults, m.body)],
            [],
            [],
        )
        return if_stmt



from __future__ import annotations

from importlib.resources import Package
from typing import TYPE_CHECKING, Any, Protocol

import pluggy

from hexdoc.utils import ValidationContext

from .book_plugin import BookPlugin
from .mod_plugin import ModPlugin
from .types import HookReturn, HookReturns

if TYPE_CHECKING:
    from hexdoc.core import Properties, ResourceLocation
    from hexdoc.minecraft import I18n
    from hexdoc.patchouli import FormatTree

HEXDOC_PROJECT_NAME = "hexdoc"

hookspec = pluggy.HookspecMarker(HEXDOC_PROJECT_NAME)


class PluginSpec(Protocol):
    @staticmethod
    @hookspec
    def hexdoc_book_plugin() -> HookReturns[BookPlugin[Any]]: ...

    @staticmethod
    @hookspec
    def hexdoc_mod_plugin(branch: str, props: Properties) -> HookReturns[ModPlugin]: ...

    @staticmethod
    @hookspec
    def hexdoc_validate_format_tree(
        tree: FormatTree,
        macros: dict[str, str],
        i18n: I18n,
        book_id: ResourceLocation,
        is_0_black: bool,
        link_overrides: dict[str, str],
    ) -> None: ...

    @staticmethod
    @hookspec
    def hexdoc_update_context(
        context: dict[str, Any],
    ) -> HookReturns[ValidationContext]: ...

    @staticmethod
    @hookspec
    def hexdoc_update_template_args(template_args: dict[str, Any]) -> None: ...

    @staticmethod
    @hookspec
    def hexdoc_load_tagged_unions() -> HookReturns[Package]: ...


# mmmmmm, interfaces


class PluginImpl(Protocol):
    """Interface for an implementation of a hexdoc plugin hook.

    These protocols are optional - they gives better type checking, but everything will
    work fine with a standard pluggy hook implementation.
    """


class BookPluginImpl(PluginImpl, Protocol):
    @staticmethod
    def hexdoc_book_plugin() -> HookReturn[BookPlugin[Any]]:
        """If your plugin represents a book system (like Patchouli), this must return an
        instance of a subclass of `BookPlugin` with all abstract methods implemented."""
        ...


class ModPluginImpl(PluginImpl, Protocol):
    @staticmethod
    def hexdoc_mod_plugin(branch: str) -> HookReturn[ModPlugin]:
        """If your plugin represents a Minecraft mod, this must return an instance of a
        subclass of `ModPlugin` or `ModPluginWithBook`, with all abstract methods
        implemented."""
        ...


class ModPluginImplWithProps(PluginImpl, Protocol):
    @staticmethod
    def hexdoc_mod_plugin(branch: str, props: Properties) -> HookReturn[ModPlugin]:
        """If your plugin represents a Minecraft mod, this must return an instance of a
        subclass of `ModPlugin` or `ModPluginWithBook`, with all abstract methods
        implemented."""
        ...


class ValidateFormatTreeImpl(PluginImpl, Protocol):
    @staticmethod
    def hexdoc_validate_format_tree(
        tree: FormatTree,
        macros: dict[str, str],
        book_id: ResourceLocation,
        i18n: I18n,
        is_0_black: bool,
        link_overrides: dict[str, str],
    ) -> None:
        """This is called as the last step when a FormatTree (styled Patchouli text) is
        generated. You can use this to modify or validate the text and styles.

        For example, Hex Casting uses this to ensure all $(action) styles are in a link.
        """
        ...


class UpdateContextImpl(PluginImpl, Protocol):
    @staticmethod
    def hexdoc_update_context(context: dict[str, Any]) -> HookReturn[ValidationContext]:
        """Modify the book validation context.

        For example, Hex Casting uses this to add pattern data needed by pattern pages.
        """
        ...


class UpdateTemplateArgsImpl(PluginImpl, Protocol):
    @staticmethod
    def hexdoc_update_template_args(template_args: dict[str, Any]) -> None:
        """Add extra template args (global variables for the Jinja templates)."""
        ...


class LoadTaggedUnionsImpl(PluginImpl, Protocol):
    @staticmethod
    def hexdoc_load_tagged_unions() -> HookReturn[Package]:
        """Return the module(s) which contain your plugin's tagged union subtypes."""
        ...



"""Custom excpetions for favicons."""

# Standard Library
import json as _json
from pathlib import Path
from typing import Any, Dict, Generator, Iterable, Union

# Project
from ._constants import SUPPORTED_FORMATS


class FaviconsError(Exception):
    """Raise an error while running favicons."""

    def __init__(self, message: str, *args: Any, **kwargs: Any):
        """Favicons Base Exception."""
        self._message = message
        self._args = args
        self._kwargs = kwargs

    @property
    def message(self) -> str:
        """Format message with args & kwargs."""
        return self._message.format(*self._args, **self._kwargs)

    @property
    def kwargs(self) -> Dict:
        """Keyword arguments as a dict."""
        return dict(self._kwargs)

    def __repr__(self) -> str:
        """Representation of exception."""
        attrs = (
            f"message='{self.message}'",
            *(repr(a) for a in self.args),
            *(f"{k}={repr(v)}" for k, v in self.kwargs.items()),
        )

        return f"{self.__class__.__name__}({', '.join(attrs)})"

    def __str__(self) -> str:
        """Represent exception as string."""
        return self.message

    def dict(self) -> Dict:
        """Represent exception as dict."""
        return {
            "message": self.message,
            "arguments": list(self.args),
            "keyword_arguments": self.kwargs,
        }

    def json(self) -> str:
        """Represent exception as JSON string."""
        return _json.dumps(self.dict())


class FaviconNotFoundError(FaviconsError):
    """Raised if the source favicon doesn't exist."""

    def __init__(self, file: Path) -> None:
        """Set message."""
        super().__init__("{} does not exit.", str(file))


class FaviconNotSupportedError(FaviconsError):
    """Raised if the source favicon file format is not supported."""

    def __init__(self, file: Path) -> None:
        """Set message."""
        one_of = ", ".join(f"'{f}'" for f in SUPPORTED_FORMATS)
        super().__init__(
            "Extension {extension} is not supported ({file}). Must be one of {one_of}.",
            extension=file.suffix,
            file=str(file),
            one_of=one_of,
        )


class FaviconColorError(FaviconsError):
    """Raised if an input color is invalid."""

    def __init__(
        self,
        color: Union[str, Iterable, Generator],
        message: str = "Color '{color}' is not a valid color.",
    ) -> None:
        """Set message."""
        if isinstance(color, Iterable):
            color = str(color)
        elif isinstance(color, Generator):
            color = ",".join(color)
        super().__init__(message, color=color)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Self, dataclass_transform

from pydantic import GetJsonSchemaHandler, TypeAdapter, ValidationInfo, model_validator
from pydantic.functional_validators import ModelWrapValidatorHandler
from pydantic_core import core_schema as cs

from hexdoc.core.resource import (
    BaseResourceLocation,
    ItemStack,
    ResourceLocation,
)

from .base import HexdocModel


class BaseInlineModel(HexdocModel):
    @classmethod
    @abstractmethod
    def _id_type(cls) -> type[BaseResourceLocation]: ...

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: cs.CoreSchema,
        handler: GetJsonSchemaHandler,
    ) -> dict[str, Any]:
        return handler(TypeAdapter(cls._id_type()).core_schema)


@dataclass_transform()
class InlineModel(BaseInlineModel, ABC):
    @classmethod
    def _id_type(cls):
        return ResourceLocation

    @classmethod
    @abstractmethod
    def load_id(cls, id: ResourceLocation, context: dict[str, Any]) -> Any: ...

    @model_validator(mode="wrap")
    @classmethod
    def _wrap_root_load_from_id(
        cls,
        value: Any,
        handler: ModelWrapValidatorHandler[Self],
        info: ValidationInfo,
    ) -> Self:
        """Loads the recipe from json if the actual value is a resource location str.

        Uses a wrap validator so we load the file *before* resolving the tagged union.
        """
        # if necessary, convert the id to a ResourceLocation
        match value:
            case str():
                id = ResourceLocation.from_str(value)
            case ResourceLocation() as id:
                pass
            case _:
                return handler(value)

        # load the data
        assert info.context is not None
        return handler(cls.load_id(id, info.context))


@dataclass_transform()
class InlineItemModel(BaseInlineModel, ABC):
    @classmethod
    def _id_type(cls):
        return ItemStack

    @classmethod
    @abstractmethod
    def load_id(cls, item: ItemStack, context: dict[str, Any]) -> Any: ...

    @model_validator(mode="wrap")
    @classmethod
    def _wrap_root_load_from_id(
        cls,
        value: Any,
        handler: ModelWrapValidatorHandler[Self],
        info: ValidationInfo,
    ) -> Self:
        """Loads the recipe from json if the actual value is a resource location str.

        Uses a wrap validator so we load the file *before* resolving the tagged union.
        """
        # if necessary, convert the id to a ItemStack
        match value:
            case str():
                item = ItemStack.from_str(value)
            case ItemStack() as item:
                pass
            case ResourceLocation() as id:
                item = ItemStack(namespace=id.namespace, path=id.path)
            case _:
                return handler(value)

        # load the data
        assert info.context is not None
        return handler(cls.load_id(item, info.context))



from __future__ import annotations

import math
import re
from abc import ABC, abstractmethod
from typing import Annotated, Literal, Self

from pydantic import AfterValidator, Field, model_validator

from hexdoc.core import ResourceLocation
from hexdoc.model import HexdocModel
from hexdoc.model.base import IGNORE_EXTRA_CONFIG
from hexdoc.utils.types import Vec3, Vec4, clamped


class BaseMinecraftModel(HexdocModel, ABC):
    """Base class for Minecraft block/item models.

    https://minecraft.wiki/w/Tutorials/Models
    """

    model_config = IGNORE_EXTRA_CONFIG

    parent: ResourceLocation | None = None
    """Loads a different model from the given path, in form of a resource location.

    If both "parent" and "elements" are set, the "elements" tag overrides the "elements"
    tag from the previous model.
    """
    display: dict[DisplayPositionName, DisplayPosition] = Field(default_factory=dict)
    """Holds the different places where item models are displayed.

    `fixed` refers to item frames, while the rest are as their name states.
    """
    textures: dict[str, TextureVariable | ResourceLocation] = Field(
        default_factory=dict
    )
    """Holds the textures of the model, in form of a resource location or can be another
    texture variable."""
    elements: list[ModelElement] | None = None
    """Contains all the elements of the model. They can have only cubic forms.

    If both "parent" and "elements" are set, the "elements" tag overrides the "elements"
    tag from the previous model.
    """
    gui_light: Literal["front", "side"] = Field(None, validate_default=False)  # type: ignore
    """If set to `side` (default), the model is rendered like a block.

    If set to `front`, model is shaded like a flat item.

    Note: although the wiki only lists this field for item models, Minecraft sets it in
    the models `minecraft:block/block` and `minecraft:block/calibrated_sculk_sensor`.
    """

    @abstractmethod
    def apply_parent(self, parent: Self):
        """Merge the parent model into this one."""
        self.parent = parent.parent
        # prefer current display/textures over parent
        self.display = parent.display | self.display
        self.textures = parent.textures | self.textures
        # only use parent elements if current model doesn't have elements
        if self.elements is None:
            self.elements = parent.elements
        if not self._was_gui_light_set:
            self.gui_light = parent.gui_light

    @model_validator(mode="after")
    def _set_default_gui_light(self):
        self._was_gui_light_set = self.gui_light is not None  # type: ignore
        if not self._was_gui_light_set:
            self.gui_light = "side"
        return self


def _validate_texture_variable(value: str):
    assert re.fullmatch(r"#\w+", value)
    return value


TextureVariable = Annotated[str, AfterValidator(_validate_texture_variable)]


DisplayPositionName = Literal[
    "thirdperson_righthand",
    "thirdperson_lefthand",
    "firstperson_righthand",
    "firstperson_lefthand",
    "gui",
    "head",
    "ground",
    "fixed",
]
"""`fixed` refers to item frames, while the rest are as their name states."""


class DisplayPosition(HexdocModel):
    """Place where an item model is displayed. Holds its rotation, translation and scale
    for the specified situation.

    Note that translations are applied to the model before rotations.

    If this is specified but not all of translation, rotation and scale are in it, the
    others aren't inherited from the parent. (TODO: is this only for items? see wiki)

    https://minecraft.wiki/w/Tutorials/Models
    """

    rotation: Vec3 = (0, 0, 0)
    """Specifies the rotation of the model according to the scheme [x, y, z]."""
    translation: Vec3[Annotated[float, clamped(-80, 80)]] = (0, 0, 0)
    """Specifies the position of the model according to the scheme [x, y, z].

    The values are clamped between -80 and 80.
    """
    scale: Vec3[Annotated[float, clamped(None, 4), Field(ge=0)]] = Field((0, 0, 0))
    """Specifies the scale of the model according to the scheme [x, y, z].

    If the value is greater than 4, it is displayed as 4.
    """

    @property
    def eulers(self) -> Vec3:
        """Euler rotation vector, in radians."""
        rotation = tuple(math.radians(v) for v in self.rotation)
        assert len(rotation) == 3
        return rotation


class ModelElement(HexdocModel):
    """An element of a block/item model. Must be cubic.

    https://minecraft.wiki/w/Tutorials/Models
    """

    model_config = IGNORE_EXTRA_CONFIG

    from_: Vec3[Annotated[float, clamped(-16, 32)]] = Field(alias="from")
    """Start point of a cuboid according to the scheme [x, y, z].

    Values must be between -16 and 32.
    """
    to: Vec3[Annotated[float, clamped(-16, 32)]]
    """Stop point of a cuboid according to the scheme [x, y, z].

    Values must be between -16 and 32.
    """
    rotation: ElementRotation | None = None
    """Defines the rotation of an element."""
    shade: bool = True
    """Defines if shadows are rendered."""
    faces: dict[FaceName, ElementFace]
    """Holds all the faces of the cuboid. If a face is left out, it does not render."""


class ElementRotation(HexdocModel):
    """Defines the rotation of an element.

    https://minecraft.wiki/w/Tutorials/Models
    """

    origin: Vec3
    """Sets the center of the rotation according to the scheme [x, y, z]."""
    axis: Literal["x", "y", "z"]
    """Specifies the direction of rotation."""
    angle: float = Field(ge=-45, le=45, multiple_of=22.5)
    """Specifies the angle of rotation.

    Can be 45 through -45 degrees in 22.5 degree increments.
    """
    rescale: bool = False
    """Specifies whether or not to scale the faces across the whole block.

    (TODO: implement)
    """

    @property
    def eulers(self) -> Vec3:
        """Euler rotation vector, in radians."""
        angle = math.radians(self.angle)
        match self.axis:
            case "x":
                return (angle, 0, 0)
            case "y":
                return (0, angle, 0)
            case "z":
                return (0, 0, angle)


FaceName = Literal["down", "up", "north", "south", "west", "east"]


class ElementFace(HexdocModel):
    """Contains the properties of the specified cuboid face.

    https://minecraft.wiki/w/Tutorials/Models
    """

    raw_uv: Vec4[Annotated[float, Field(ge=0, le=16)]] | None = Field(None, alias="uv")
    """Defines the area of the texture to use according to the scheme [x1, y1, x2, y2].

    The texture behavior is inconsistent if UV extends below 0 or above 16, so these
    values are forbidden by field validation.

    If the numbers of x1 and x2 are swapped (e.g. from 0, 0, 16, 16 to 16, 0, 0, 16),
    the texture flips.

    UV is optional, and if not supplied it automatically generates based on the
    element's position.
    """
    texture: TextureVariable
    """Specifies the texture in form of the texture variable prepended with a #."""
    cullface: FaceName | None = None
    """Specifies whether a face does not need to be rendered when there is a block
    touching it in the specified position.

    It also determines the side of the block to use the light level from for lighting
    the face, and if unset, defaults to the side.

    (TODO: implement; not sure what this means)
    """
    rotation: Literal[0, 90, 180, 270] = 0
    """Rotates the texture clockwise by the specified number of degrees.

    Rotation does not affect which part of the texture is used. Instead, it amounts to
    permutation of the selected texture vertexes (selected implicitly, or explicitly
    though uv).
    """
    tintindex: int = -1
    """Determines whether to tint the texture using a hardcoded tint index.

    The default value, -1, indicates not to use the tint. Any other number is provided
    to BlockColors to get the tint value corresponding to that index. However, most
    blocks do not have a tint value defined (in which case white is used). Furthermore,
    no vanilla block currently uses multiple tint values, and thus the tint index value
    is ignored (as long as it is set to something other than -1); it could be used for
    modded blocks that need multiple distinct tint values in the same block though.

    (TODO: implement)
    """

    @property
    def uv(self):
        if self.raw_uv is None:
            return None
        return ElementFaceUV(uvs=self.raw_uv, rotation=self.rotation)

    @property
    def texture_name(self):
        """Returns `self.texture` without the leading `#`."""
        return self.texture.lstrip("#")


class ElementFaceUV(HexdocModel):
    uvs: Vec4[Annotated[float, Field(ge=0, le=16)]]
    rotation: Literal[0, 90, 180, 270] = 0

    @classmethod
    def default(cls, element: ModelElement, direction: FaceName):
        x1, y1, z1 = element.from_
        x2, y2, z2 = element.to

        uvs: Vec4
        match direction:
            case "down":
                uvs = (x1, 16 - z2, x2, 16 - z1)
            case "up":
                uvs = (x1, z1, x2, z2)
            case "north":
                uvs = (16 - x2, 16 - y2, 16 - x1, 16 - y1)
            case "south":
                uvs = (x1, 16 - y2, x2, 16 - y1)
            case "west":
                uvs = (z1, 16 - y2, z2, 16 - y1)
            case "east":
                uvs = (16 - z2, 16 - y2, 16 - z1, 16 - y1)

        return cls(uvs=uvs)

    def get_uv(self, index: Literal[0, 1, 2, 3]):
        return self.get_u(index), self.get_v(index)

    def get_u(self, index: Literal[0, 1, 2, 3]):
        if self._get_shifted_index(index) in {0, 1}:
            return self.uvs[0]
        return self.uvs[2]

    def get_v(self, index: Literal[0, 1, 2, 3]):
        if self._get_shifted_index(index) in {0, 3}:
            return self.uvs[1]
        return self.uvs[3]

    def _get_shifted_index(self, index: Literal[0, 1, 2, 3]):
        return (index + self.rotation // 90) % 4


# this is required to ensure BlockModel and ItemModel are fully defined
BaseMinecraftModel.model_rebuild()



from __future__ import annotations

from typing import Annotated, Any, Literal

from frozendict import frozendict
from pydantic import AfterValidator, BeforeValidator, ConfigDict, model_validator

from hexdoc.core import ResourceLocation
from hexdoc.model import HexdocModel
from hexdoc.model.base import DEFAULT_CONFIG


class Blockstate(HexdocModel):
    """Represents a Minecraft blockstate file.

    There are several different variants of some blocks (like doors, which can be open
    or closed), hence each block has its own block state file, which lists all its
    existing variants and links them to their corresponding models. Blocks can also be
    composed of several different models at the same time, called "multipart". The
    models are then used depending on the block states of the block.

    These files are stored in the following folder: `assets/<namespace>/blockstates`.
    The files are used directly based on their filename, thus a block state file with
    another name than the existing ones does not affect any block.

    https://minecraft.wiki/w/Tutorials/Models#Block_states
    """

    model_config = DEFAULT_CONFIG | ConfigDict(
        arbitrary_types_allowed=True,
    )

    variants: dict[_BlockstateVariantKey, _BlockstateModels] | None = None
    """Holds the names of all the variants of the block.

    Name of a variant, which consists of the relevant block states separated by commas.
    A block with just one variant uses "" as a name for its variant. Each variant can
    have one model or an array of models and contains their properties. If set to an
    array, the model is chosen randomly from the options given.
    """
    multipart: list[BlockstateMultipart] | None = None
    """Used instead of variants to combine models based on block state attributes."""

    @model_validator(mode="after")
    def _validate_variant_or_multipart(self):
        match bool(self.variants), bool(self.multipart):
            case True, True:
                reason = "both"
            case False, False:
                reason = "neither"
            case _:
                return self
        raise ValueError(f"Expected exactly one of variants or multipart, got {reason}")


class BlockstateModel(HexdocModel):
    """Contains the properties of a model.

    https://minecraft.wiki/w/Tutorials/Models#Block_states
    """

    model: ResourceLocation
    """Specifies the path to the model file of the block."""
    x: Literal[0, 90, 180, 270] = 0
    """Rotation of the model on the x-axis in increments of 90 degrees."""
    y: Literal[0, 90, 180, 270] = 0
    """Rotation of the model on the y-axis in increments of 90 degrees."""
    uvlock: bool = False
    """Locks the rotation of the texture of a block, if set to true. This way the
    texture does not rotate with the block when using the x and y-tags above."""
    weight: int = 1
    """Sets the probability of the model for being used in the game, defaults to 100%.

    If more than one model is used for the same variant, the probability is calculated
    by dividing the individual model's weight by the sum of the weights of all models.
    """


class BlockstateMultipart(HexdocModel):
    """Determines a case and the model that should apply in that case.

    https://minecraft.wiki/w/Tutorials/Models#Block_states
    """

    apply: _BlockstateModels
    """Determines the model(s) to apply and its properties.

    If set to an array, the model is chosen randomly from the options given.
    """
    when: _MultipartCondition | None = None
    """A list of cases that have to be met for the model to be applied.

    If unset, the model always applies.
    """


def _validate_variant_key(value: Any):
    if isinstance(value, str):
        value = dict(item.split("=") for item in value.split(","))
    return frozendict[Any, Any](value)


_BlockstateVariantKey = Annotated[
    frozendict[str, str],
    BeforeValidator(_validate_variant_key),
]
"""Name of a variant, which consists of the relevant block states separated by commas.

A block with just one variant uses "" as a name for its variant.

Item frames are treated as blocks and use "map=false" for a map-less item frame, and
"map=true" for item frames with maps.
"""


_BlockstateModels = BlockstateModel | list[BlockstateModel]
"""There can be one model or an array of models.

If set to an array, the model is chosen randomly from the options given.
"""


def _validate_states_key(value: str):
    assert value not in {"AND", "OR"}, "State name cannot be AND/OR"
    return value


def _validate_states_value(value: Any):
    if isinstance(value, str):
        return set(value.split("|"))
    return value


_MultipartStates = dict[
    Annotated[str, AfterValidator(_validate_states_key)],
    Annotated[set[str], BeforeValidator(_validate_states_value)],
]
"""A list of cases that all have to match the block to return true.

Name of a block state. A single case that has to match one of the block states. It can
be set to a list separated by | to allow multiple values to match.
"""

_MultipartCondition = (
    dict[Literal["AND"], list[_MultipartStates]]
    | dict[Literal["OR"], list[_MultipartStates]]
    | _MultipartStates
)
"""A list of cases that have to be met for the model to be applied.

OR: Matches if any of the contained cases return true. Cannot be set alongside other
cases.

AND: Matches if all of the contained cases return true. Cannot be set alongside other
cases.
"""



__all__ = [
    "FaviconColorError",
    "FaviconNotFoundError",
    "FaviconNotSupportedError",
    "Favicons",
    "FaviconsError",
]

from ._exceptions import (
    FaviconColorError,
    FaviconNotFoundError,
    FaviconNotSupportedError,
    FaviconsError,
)
from ._generate import Favicons



# pyright: reportUnknownMemberType=false

from __future__ import annotations

import logging
import math
from dataclasses import dataclass
from enum import Flag, auto
from functools import cached_property
from pathlib import Path
from typing import Any, Literal, cast

import importlib_resources as resources
import moderngl as mgl
import moderngl_window as mglw
import numpy as np
from moderngl import Context, Program, Uniform
from moderngl_window import WindowConfig
from moderngl_window.context.headless import Window as HeadlessWindow
from moderngl_window.opengl.vao import VAO
from PIL import Image
from pydantic import ValidationError
from pyrr import Matrix44

from hexdoc.core import ModResourceLoader, ResourceLocation
from hexdoc.graphics import glsl
from hexdoc.minecraft.assets import AnimationMeta
from hexdoc.minecraft.assets.animated import AnimationMetaTag
from hexdoc.minecraft.models import BlockModel
from hexdoc.minecraft.models.base_model import (
    DisplayPosition,
    ElementFace,
    ElementFaceUV,
    FaceName,
    ModelElement,
)
from hexdoc.utils.types import Vec3, Vec4

logger = logging.getLogger(__name__)


# https://minecraft.wiki/w/Help:Isometric_renders#Preferences
LIGHT_TOP = 0.98
LIGHT_LEFT = 0.8
LIGHT_RIGHT = 0.608

LIGHT_FLAT = 0.98


class DebugType(Flag):
    NONE = 0
    AXES = auto()
    NORMALS = auto()


@dataclass(kw_only=True)
class BlockRenderer:
    loader: ModResourceLoader
    output_dir: Path | None = None
    debug: DebugType = DebugType.NONE

    def __post_init__(self):
        self.window = HeadlessWindow(
            size=(300, 300),
        )
        mglw.activate_context(self.window)

        self.config = BlockRendererConfig(ctx=self.window.ctx, wnd=self.window)

        self.window.config = self.config
        self.window.swap_buffers()
        self.window.set_default_viewport()

    @property
    def ctx(self):
        return self.window.ctx

    def render_block_model(
        self,
        model: BlockModel | ResourceLocation,
        output_path: str | Path,
    ):
        if isinstance(model, ResourceLocation):
            _, model = self.loader.load_resource(
                type="assets",
                folder="models",
                id=model,
                decode=BlockModel.model_validate_json,
            )

        model.load_parents_and_apply(self.loader)

        textures = {
            name: self.load_texture(texture_id)
            for name, texture_id in model.resolve_texture_variables().items()
        }

        output_path = Path(output_path)
        if self.output_dir and not output_path.is_absolute():
            output_path = self.output_dir / output_path

        self.config.render_block(model, textures, output_path, self.debug)

    def load_texture(self, texture_id: ResourceLocation):
        logger.debug(f"Loading texture: {texture_id}")
        _, path = self.loader.find_resource("assets", "textures", texture_id + ".png")

        meta_path = path.with_suffix(".png.mcmeta")
        if meta_path.is_file():
            logger.debug(f"Loading animation mcmeta: {meta_path}")
            # FIXME: hack
            try:
                meta = AnimationMeta.model_validate_json(meta_path.read_bytes())
            except ValidationError as e:
                logger.warning(f"Failed to parse animation meta for {texture_id}:\n{e}")
                meta = None
        else:
            meta = None

        return BlockTextureInfo(path, meta)

    def destroy(self):
        self.window.destroy()

    def __enter__(self):
        return self

    def __exit__(self, *_: Any):
        self.destroy()
        return False


class BlockRendererConfig(WindowConfig):
    def __init__(self, ctx: Context, wnd: HeadlessWindow):
        super().__init__(ctx, wnd)

        # depth test: ensure faces are displayed in the correct order
        # blend: handle translucency
        # cull face: remove back faces, eg. for trapdoors
        self.ctx.enable(mgl.DEPTH_TEST | mgl.BLEND | mgl.CULL_FACE)

        view_size = 16
        self.projection = Matrix44.orthogonal_projection(
            left=-view_size / 2,
            right=view_size / 2,
            top=view_size / 2,
            bottom=-view_size / 2,
            near=0.01,
            far=20_000,
            dtype="f4",
        ) * Matrix44.from_scale((1, -1, 1), "f4")

        self.camera, self.eye = direction_camera(pos="south")

        self.lights = [
            ((0, -1, 0), LIGHT_TOP),
            ((1, 0, -1), LIGHT_LEFT),
            ((-1, 0, -1), LIGHT_RIGHT),
        ]

        # block faces

        self.face_prog = self.ctx.program(
            vertex_shader=read_shader("block_face", "vertex"),
            fragment_shader=read_shader("block_face", "fragment"),
        )

        self.uniform("m_proj").write(self.projection)
        self.uniform("m_camera").write(self.camera)
        self.uniform("layer").value = 0  # TODO: implement animations

        for i, (direction, diffuse) in enumerate(self.lights):
            self.uniform(f"lights[{i}].direction").value = direction
            self.uniform(f"lights[{i}].diffuse").value = diffuse

        # axis planes

        self.debug_plane_prog = self.ctx.program(
            vertex_shader=read_shader("debug/plane", "vertex"),
            fragment_shader=read_shader("debug/plane", "fragment"),
        )

        self.uniform("m_proj", self.debug_plane_prog).write(self.projection)
        self.uniform("m_camera", self.debug_plane_prog).write(self.camera)
        self.uniform("m_model", self.debug_plane_prog).write(Matrix44.identity("f4"))

        self.debug_axes = list[tuple[VAO, Vec4]]()

        pos = 8
        neg = 0
        for from_, to, color, direction in [
            ((0, neg, neg), (0, pos, pos), (1, 0, 0, 0.75), "east"),
            ((neg, 0, neg), (pos, 0, pos), (0, 1, 0, 0.75), "up"),
            ((neg, neg, 0), (pos, pos, 0), (0, 0, 1, 0.75), "south"),
        ]:
            vao = VAO()
            verts = get_face_verts(from_, to, direction)
            vao.buffer(np.array(verts, np.float32), "3f", ["in_position"])
            self.debug_axes.append((vao, color))

        # vertex normal vectors

        self.debug_normal_prog = self.ctx.program(
            vertex_shader=read_shader("debug/normal", "vertex"),
            geometry_shader=read_shader("debug/normal", "geometry"),
            fragment_shader=read_shader("debug/normal", "fragment"),
        )

        self.uniform("m_proj", self.debug_normal_prog).write(self.projection)
        self.uniform("m_camera", self.debug_normal_prog).write(self.camera)
        self.uniform("lineSize", self.debug_normal_prog).value = 4

        self.ctx.line_width = 3

    def render_block(
        self,
        model: BlockModel,
        texture_vars: dict[str, BlockTextureInfo],
        output_path: Path,
        debug: DebugType = DebugType.NONE,
    ):
        if not model.elements:
            raise ValueError("Unable to render model, no elements found")

        self.wnd.clear()

        # enable/disable flat item lighting
        match model.gui_light:
            case "front":
                flatLighting = LIGHT_FLAT
            case "side":
                flatLighting = 0
        self.uniform("flatLighting").value = flatLighting

        # load textures
        texture_locs = dict[str, int]()
        transparent_textures = set[str]()

        for i, (name, info) in enumerate(texture_vars.items()):
            texture_locs[name] = i

            logger.debug(f"Loading texture {name}: {info}")
            image = Image.open(info.image_path).convert("RGBA")

            extrema = image.getextrema()
            assert len(extrema) >= 4, f"Expected 4 bands but got {len(extrema)}"
            min_alpha, _ = extrema[3]
            if min_alpha < 255:
                logger.debug(f"Transparent texture: {name} ({min_alpha=})")
                transparent_textures.add(name)

            # TODO: implement non-square animations, write test cases
            match info.meta:
                case AnimationMeta(
                    animation=AnimationMetaTag(height=frame_height),
                ) if frame_height:
                    # animated with specified size
                    layers = image.height // frame_height
                case AnimationMeta():
                    # size is unspecified, assume it's square and verify later
                    frame_height = image.width
                    layers = image.height // frame_height
                case None:
                    # non-animated
                    frame_height = image.height
                    layers = 1

            if frame_height * layers != image.height:
                raise RuntimeError(
                    f"Invalid texture size for variable #{name}:"
                    + f" {frame_height}x{layers} != {image.height}"
                    + f"\n  {info}"
                )

            logger.debug(f"Texture array: {image.width=}, {frame_height=}, {layers=}")
            texture = self.ctx.texture_array(
                size=(image.width, frame_height, layers),
                components=4,
                data=image.tobytes(),
            )
            texture.filter = (mgl.NEAREST, mgl.NEAREST)
            texture.use(i)

        # transform entire model

        gui = model.display.get("gui") or DisplayPosition(
            rotation=(30, 225, 0),
            translation=(0, 0, 0),
            scale=(0.625, 0.625, 0.625),
        )

        model_transform = cast(
            Matrix44,
            Matrix44.from_scale(gui.scale, "f4")
            * get_rotation_matrix(gui.eulers)
            * Matrix44.from_translation(gui.translation, "f4")
            * Matrix44.from_translation((-8, -8, -8), "f4"),
        )

        normals_transform = Matrix44.from_y_rotation(-gui.eulers[1], "f4")
        self.uniform("m_normals").write(normals_transform)

        # render elements

        baked_faces = list[BakedFace]()

        for element in model.elements:
            element_transform = model_transform.copy()

            # TODO: rescale??
            if rotation := element.rotation:
                origin = np.array(rotation.origin)
                element_transform *= cast(
                    Matrix44,
                    Matrix44.from_translation(origin, "f4")
                    * get_rotation_matrix(rotation.eulers)
                    * Matrix44.from_translation(-origin, "f4"),
                )

            # prepare each face of the element for rendering
            for direction, face in element.faces.items():
                baked_face = BakedFace(
                    element=element,
                    direction=direction,
                    face=face,
                    m_model=element_transform,
                    texture0=texture_locs[face.texture_name],
                    is_opaque=face.texture_name not in transparent_textures,
                )
                baked_faces.append(baked_face)

        # TODO: use a map if this is actually slow
        baked_faces.sort(key=lambda face: face.sortkey(self.eye))

        for face in baked_faces:
            self.uniform("m_model").write(face.m_model)
            self.uniform("texture0").value = face.texture0

            face.vao.render(self.face_prog)

            if DebugType.NORMALS in debug:
                self.uniform("m_model", self.debug_normal_prog).write(face.m_model)
                face.vao.render(self.debug_normal_prog)

        if DebugType.AXES in debug:
            self.ctx.disable(mgl.CULL_FACE)
            for axis, color in self.debug_axes:
                self.uniform("color", self.debug_plane_prog).value = color
                axis.render(self.debug_plane_prog)
            self.ctx.enable(mgl.CULL_FACE)

        self.ctx.finish()

        # save to file

        image = Image.frombytes(
            mode="RGBA",
            size=self.wnd.fbo.size,
            data=self.wnd.fbo.read(components=4),
        ).transpose(Image.Transpose.FLIP_TOP_BOTTOM)

        output_path.parent.mkdir(parents=True, exist_ok=True)
        image.save(output_path, format="png")

    def uniform(self, name: str, program: Program | None = None):
        program = program or self.face_prog
        assert isinstance(uniform := program[name], Uniform)
        return uniform


@dataclass
class BlockTextureInfo:
    image_path: Path
    meta: AnimationMeta | None


@dataclass(kw_only=True)
class BakedFace:
    element: ModelElement
    direction: FaceName
    face: ElementFace
    m_model: Matrix44
    texture0: float
    is_opaque: bool

    def __post_init__(self):
        self.verts = get_face_verts(self.element.from_, self.element.to, self.direction)

        self.normals = get_face_normals(self.direction)

        face_uv = self.face.uv or ElementFaceUV.default(self.element, self.direction)
        self.uvs = [
            value
            for index in get_face_uv_indices(self.direction)
            for value in face_uv.get_uv(index)
        ]

        self.vao = VAO()
        self.vao.buffer(np.array(self.verts, np.float32), "3f", ["in_position"])
        self.vao.buffer(np.array(self.normals, np.float32), "3f", ["in_normal"])
        self.vao.buffer(np.array(self.uvs, np.float32) / 16, "2f", ["in_texcoord_0"])

    @cached_property
    def position(self):
        x, y, z, n = 0, 0, 0, 0
        for i in range(0, len(self.verts), 3):
            x += self.verts[i]
            y += self.verts[i + 1]
            z += self.verts[i + 2]
            n += 1
        return (x / n, y / n, z / n)

    def sortkey(self, eye: Vec3):
        if self.is_opaque:
            return 0
        return sum((a - b) ** 2 for a, b in zip(eye, self.position))


def get_face_verts(from_: Vec3, to: Vec3, direction: FaceName):
    x1, y1, z1 = from_
    x2, y2, z2 = to

    # fmt: off
    match direction:
        case "south":
            return [
                x2, y1, z2,
                x2, y2, z2,
                x1, y1, z2,
                x2, y2, z2,
                x1, y2, z2,
                x1, y1, z2,
            ]
        case "east":
            return [
                x2, y1, z1,
                x2, y2, z1,
                x2, y1, z2,
                x2, y2, z1,
                x2, y2, z2,
                x2, y1, z2,
            ]
        case "down":
            return [
                x2, y1, z1,
                x2, y1, z2,
                x1, y1, z2,
                x2, y1, z1,
                x1, y1, z2,
                x1, y1, z1,
            ]
        case "west":
            return [
                x1, y1, z2,
                x1, y2, z2,
                x1, y2, z1,
                x1, y1, z2,
                x1, y2, z1,
                x1, y1, z1,
            ]
        case "north":
            return [
                x2, y2, z1,
                x2, y1, z1,
                x1, y1, z1,
                x2, y2, z1,
                x1, y1, z1,
                x1, y2, z1,
            ]
        case "up":
            return [
                x2, y2, z1,
                x1, y2, z1,
                x2, y2, z2,
                x1, y2, z1,
                x1, y2, z2,
                x2, y2, z2,
            ]
    # fmt: on


def get_face_normals(direction: FaceName):
    return 6 * get_direction_vec(direction)


def get_face_uv_indices(direction: FaceName):
    match direction:
        case "south":
            return (2, 3, 1, 3, 0, 1)
        case "east":
            return (2, 3, 1, 3, 0, 1)
        case "down":
            return (2, 3, 0, 2, 0, 1)
        case "west":
            return (2, 3, 0, 2, 0, 1)
        case "north":
            return (0, 1, 2, 0, 2, 3)
        case "up":
            return (3, 0, 2, 0, 1, 2)


def orbit_camera(pitch: float, yaw: float):
    """Both values are in degrees."""

    eye = transform_vec(
        (-64, 0, 0),
        cast(
            Matrix44,
            Matrix44.identity(dtype="f4")
            * Matrix44.from_y_rotation(math.radians(yaw))
            * Matrix44.from_z_rotation(math.radians(pitch)),
        ),
    )

    up = transform_vec(
        (-1, 0, 0),
        cast(
            Matrix44,
            Matrix44.identity(dtype="f4")
            * Matrix44.from_y_rotation(math.radians(yaw))
            * Matrix44.from_z_rotation(math.radians(90 - pitch)),
        ),
    )

    return Matrix44.look_at(
        eye=eye,
        target=(0, 0, 0),
        up=up,
        dtype="f4",
    ), eye


def transform_vec(vec: Vec3, matrix: Matrix44) -> Vec3:
    return np.matmul((*vec, 1), matrix, dtype="f4")[:3]


def direction_camera(pos: FaceName, up: FaceName = "up"):
    """eg. north -> camera is placed to the north of the model, looking south"""
    eye = get_direction_vec(pos, 64)
    return Matrix44.look_at(
        eye=eye,
        target=(0, 0, 0),
        up=get_direction_vec(up),
        dtype="f4",
    ), eye


def get_direction_vec(direction: FaceName, magnitude: float = 1):
    match direction:
        case "north":
            return (0, 0, -magnitude)
        case "south":
            return (0, 0, magnitude)
        case "west":
            return (-magnitude, 0, 0)
        case "east":
            return (magnitude, 0, 0)
        case "down":
            return (0, -magnitude, 0)
        case "up":
            return (0, magnitude, 0)


def read_shader(path: str, type: Literal["fragment", "vertex", "geometry"]):
    file = resources.files(glsl) / path / f"{type}.glsl"
    return file.read_text("utf-8")


def get_rotation_matrix(eulers: Vec3) -> Matrix44:
    return cast(
        Matrix44,
        Matrix44.from_x_rotation(-eulers[0], "f4")
        * Matrix44.from_y_rotation(-eulers[1], "f4")
        * Matrix44.from_z_rotation(-eulers[2], "f4"),
    )



import logging
from typing import Any, Self

from pydantic import Field
from pydantic.json_schema import SkipJsonSchema

from hexdoc.core import ResourceLocation
from hexdoc.core.loader import ModResourceLoader
from hexdoc.minecraft import LocalizedStr
from hexdoc.minecraft.assets import ItemWithTexture, NamedTexture
from hexdoc.model import IDModel
from hexdoc.utils import Sortable, sorted_dict
from hexdoc.utils.graphs import TypedDiGraph

from .entry import Entry
from .text import FormatTree
from .utils import Flagged

logger = logging.getLogger(__name__)


class Category(IDModel, Sortable, Flagged):
    """Category with pages and localizations.

    See: https://vazkiimods.github.io/Patchouli/docs/reference/category-json
    """

    entries: SkipJsonSchema[dict[ResourceLocation, Entry]] = Field(default_factory=dict)
    is_spoiler: SkipJsonSchema[bool] = False

    # required
    name: LocalizedStr
    description: FormatTree
    icon: ItemWithTexture | NamedTexture

    # optional
    parent_id: ResourceLocation | None = Field(default=None, alias="parent")
    _parent_cmp_key: tuple[int, ...] | None = None
    sortnum: int = 0
    secret: bool = False

    @classmethod
    def load_all(
        cls,
        context: dict[str, Any],
        book_id: ResourceLocation,
        use_resource_pack: bool,
    ) -> dict[ResourceLocation, Self]:
        loader = ModResourceLoader.of(context)

        # load
        categories = dict[ResourceLocation, Self]()
        G = TypedDiGraph[ResourceLocation]()

        for resource_dir, id, data in loader.load_book_assets(
            book_id,
            "categories",
            use_resource_pack,
        ):
            # Patchouli checks flags before resolving category parents
            # https://github.com/VazkiiMods/Patchouli/blob/abd6d03a08c37bcf116730021fda9f477412b31f/Xplat/src/main/java/vazkii/patchouli/client/book/BookContentsBuilder.java#L151
            category = cls.load(resource_dir, id, data, context)
            if not category.is_flag_enabled:
                logger.info(
                    f"Skipping category {id} due to disabled flag {category.flag}"
                )
                continue

            categories[id] = category
            if category.parent_id:
                G.add_edge(category.parent_id, category.id)

        # if there's a cycle in the graph, we can't find a valid ordering
        # eg. two categories with each other as parents
        if cycle := G.find_cycle():
            raise ValueError(
                "Found cycle of category parents:\n  "
                + "\n  ".join(f"{u} -> {v}" for u, v in cycle)
            )

        # late-init _parent_cmp_key
        for parent_id in G.topological_sort():
            parent = categories.get(parent_id)
            if parent is None:
                children = ", ".join(str(v) for _, v in G.iter_out_edges(parent_id))
                raise ValueError(
                    f"Parent category {parent_id} required by {children} does not exist"
                )

            for _, child_id in G.iter_out_edges(parent_id):
                categories[child_id]._parent_cmp_key = parent._cmp_key

        # return sorted by sortnum, which requires parent to be initialized
        return sorted_dict(categories)

    @property
    def book_link_key(self):
        """Key to look up this category in `BookContext.book_links`."""
        return str(self.id)

    @property
    def fragment(self):
        """URL fragment for this category in `BookContext.book_links`."""
        return self.id.path

    @property
    def redirect_path(self):
        """Path to this category when generating redirect pages."""
        return self.id.path

    @property
    def _is_cmp_key_ready(self) -> bool:
        return self.parent_id is None or self._parent_cmp_key is not None

    @property
    def _cmp_key(self) -> tuple[int, ...]:
        # implement Sortable
        if parent_cmp_key := self._parent_cmp_key:
            return parent_cmp_key + (self.sortnum,)
        return (self.sortnum,)



from __future__ import annotations

import importlib
from dataclasses import dataclass
from importlib.resources import Package
from pathlib import Path
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    Iterator,
    Literal,
    Mapping,
    Never,
    ParamSpec,
    Sequence,
    TypeVar,
    overload,
)

import pluggy
from jinja2 import BaseLoader, ChoiceLoader, PackageLoader
from jinja2.sandbox import SandboxedEnvironment

from hexdoc.utils import ContextSource, ValidationContext

if TYPE_CHECKING:
    from hexdoc.cli.app import LoadedBookInfo
    from hexdoc.core import Properties, ResourceLocation
    from hexdoc.minecraft import I18n
    from hexdoc.patchouli import FormatTree

import logging

from .book_plugin import BookPlugin
from .mod_plugin import DefaultRenderedTemplates, ModPlugin, ModPluginWithBook
from .specs import HEXDOC_PROJECT_NAME, PluginSpec
from .types import HookReturns

logger = logging.getLogger(__name__)

_T = TypeVar("_T")

_P = ParamSpec("_P")
_R = TypeVar("_R", covariant=True)


class PluginNotFoundError(RuntimeError):
    pass


@dataclass
class TypedHookCaller(Generic[_P, _R]):
    plugin_name: str | None
    caller: pluggy.HookCaller

    @property
    def name(self):
        return self.caller.name

    @property
    def plugin_display_name(self):
        if self.plugin_name is None:
            return HEXDOC_PROJECT_NAME

        return f"Plugin {HEXDOC_PROJECT_NAME}-{self.plugin_name}"

    def try_call(self, *args: _P.args, **kwargs: _P.kwargs) -> _R | None:
        result = self.caller(*args, **kwargs)
        match result:
            case None | []:
                return None
            case _:
                return result

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _R:
        result = self.try_call(*args, **kwargs)
        if result is None:
            raise PluginNotFoundError(
                f"{self.plugin_display_name} does not implement hook {self.name}"
            )
        return result


class _NoCallTypedHookCaller(TypedHookCaller[_P, None]):
    """Represents a TypedHookCaller which returns None. This will always raise, so the
    return type of __call__ is set to Never."""

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> Never: ...


# TODO: convert to dataclass
class PluginManager(ValidationContext):
    """Custom hexdoc plugin manager with helpers and stronger typing."""

    def __init__(self, branch: str, props: Properties, load: bool = True) -> None:
        """Initialize the hexdoc plugin manager.

        If `load` is true (the default), calls `init_entrypoints` and `init_mod_plugins`.
        """
        self.branch = branch
        self.props = props
        self.inner = pluggy.PluginManager(HEXDOC_PROJECT_NAME)
        self.mod_plugins: dict[str, ModPlugin] = {}
        self.book_plugins: dict[str, BookPlugin[Any]] = {}

        self.inner.add_hookspecs(PluginSpec)
        if load:
            self.init_entrypoints()
            self.init_plugins()

    def init_entrypoints(self):
        self.inner.load_setuptools_entrypoints(HEXDOC_PROJECT_NAME)
        self.inner.check_pending()

    def init_plugins(self):
        self._init_book_plugins()
        self._init_mod_plugins()

    def _init_book_plugins(self):
        caller = self._hook_caller(PluginSpec.hexdoc_book_plugin)
        for plugin in flatten(caller.try_call()):
            self.book_plugins[plugin.modid] = plugin

    def _init_mod_plugins(self):
        caller = self._hook_caller(PluginSpec.hexdoc_mod_plugin)
        for plugin in flatten(
            caller.try_call(
                branch=self.branch,
                props=self.props,
            )
        ):
            self.mod_plugins[plugin.modid] = plugin

    def register(self, plugin: Any, name: str | None = None):
        self.inner.register(plugin, name)
        self.init_plugins()

    def book_plugin(self, modid: str):
        plugin = self.book_plugins.get(modid)
        if plugin is None:
            raise ValueError(f"No BookPlugin registered for modid: {modid}")
        return plugin

    @overload
    def mod_plugin(self, modid: str, book: Literal[True]) -> ModPluginWithBook: ...

    @overload
    def mod_plugin(self, modid: str, book: bool = False) -> ModPlugin: ...

    def mod_plugin(self, modid: str, book: bool = False):
        plugin = self.mod_plugins.get(modid)
        if plugin is None:
            raise ValueError(f"No ModPlugin registered for modid: {modid}")

        if book and not isinstance(plugin, ModPluginWithBook):
            raise ValueError(
                f"ModPlugin registered for modid `{modid}`"
                f" does not inherit from ModPluginWithBook: {plugin}"
            )

        return plugin

    def mod_plugin_with_book(self, modid: str):
        return self.mod_plugin(modid, book=True)

    def minecraft_version(self) -> str | None:
        versions = dict[str, str]()

        for modid, plugin in self.mod_plugins.items():
            version = plugin.compat_minecraft_version
            if version is not None:
                versions[modid] = version

        match len(set(versions.values())):
            case 0:
                return None
            case 1:
                return versions.popitem()[1]
            case n:
                raise ValueError(
                    f"Got {n} Minecraft versions, expected 1: "
                    + ", ".join(
                        f"{modid}={version}" for modid, version in versions.items()
                    )
                )

    def validate_format_tree(
        self,
        tree: FormatTree,
        macros: dict[str, str],
        book_id: ResourceLocation,
        i18n: I18n,
        is_0_black: bool,
        link_overrides: dict[str, str],
    ):
        caller = self._hook_caller(PluginSpec.hexdoc_validate_format_tree)
        caller.try_call(
            tree=tree,
            macros=macros,
            book_id=book_id,
            i18n=i18n,
            is_0_black=is_0_black,
            link_overrides=link_overrides,
        )
        return tree

    def update_context(self, context: dict[str, Any]) -> Iterator[ValidationContext]:
        caller = self._hook_caller(PluginSpec.hexdoc_update_context)
        if returns := caller.try_call(context=context):
            yield from flatten(returns)

    def update_jinja_env(self, modids: Sequence[str], env: SandboxedEnvironment):
        for modid in modids:
            plugin = self.mod_plugin(modid)
            plugin.update_jinja_env(env)
        return env

    def pre_render_site(
        self,
        books: list[LoadedBookInfo],
        env: SandboxedEnvironment,
        output_dir: Path,
        modids: Sequence[str],
    ):
        for modid in modids:
            self.mod_plugin(modid).pre_render_site(self.props, books, env, output_dir)

    def post_render_site(
        self,
        books: list[LoadedBookInfo],
        env: SandboxedEnvironment,
        output_dir: Path,
        modids: Sequence[str],
    ):
        for modid in modids:
            self.mod_plugin(modid).post_render_site(self.props, books, env, output_dir)

    def update_template_args(self, template_args: dict[str, Any]):
        caller = self._hook_caller(PluginSpec.hexdoc_update_template_args)
        caller.try_call(template_args=template_args)
        return template_args

    def pre_render_book(
        self,
        template_args: dict[str, Any],
        output_dir: Path,
        modids: Sequence[str],
    ):
        for modid in modids:
            self.mod_plugin(modid).pre_render_book(template_args, output_dir)

    def post_render_book(
        self,
        template_args: dict[str, Any],
        output_dir: Path,
        modids: Sequence[str],
    ):
        for modid in modids:
            self.mod_plugin(modid).post_render_book(template_args, output_dir)

    def load_resources(self, modid: str) -> Iterator[ModuleType]:
        plugin = self.mod_plugin(modid)
        for package in flatten([plugin.resource_dirs()]):
            yield import_package(package)

    def load_tagged_unions(self) -> Iterator[ModuleType]:
        yield from self._import_from_hook(PluginSpec.hexdoc_load_tagged_unions)

    def load_jinja_templates(self, modids: Sequence[str]):
        """modid -> PackageLoader"""
        extra_modids = set(self.mod_plugins.keys()) - set(modids)

        included = self._package_loaders_for(modids)
        extra = self._package_loaders_for(extra_modids)

        return included, extra

    def _package_loaders_for(self, modids: Iterable[str]):
        loaders = dict[str, BaseLoader]()

        for modid in modids:
            plugin = self.mod_plugin(modid)

            result = plugin.jinja_template_root()
            if not result:
                continue

            loaders[modid] = ChoiceLoader(
                [
                    PackageLoader(
                        package_name=import_package(package).__name__,
                        package_path=package_path,
                    )
                    for package, package_path in flatten([result])
                ]
            )

        return loaders

    def default_rendered_templates(
        self,
        modids: Iterable[str],
        book: Any,
        context: ContextSource,
    ):
        templates: DefaultRenderedTemplates = {}
        for modid in modids:
            plugin = self.mod_plugin(modid)
            templates |= plugin.default_rendered_templates()
            templates |= plugin.default_rendered_templates_v2(book, context)

        result = dict[Path, tuple[str, Mapping[str, Any]]]()
        for path, value in templates.items():
            match value:
                case str(template):
                    args = dict[str, Any]()
                case (template, args):
                    pass
            result[Path(path)] = (template, args)

        return result

    def _import_from_hook(
        self,
        __spec: Callable[_P, HookReturns[Package]],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> Iterator[ModuleType]:
        packages = self._hook_caller(__spec)(*args, **kwargs)
        for package in flatten(packages):
            yield import_package(package)

    def load_flags(self) -> dict[str, bool]:
        # https://vazkiimods.github.io/Patchouli/docs/patchouli-basics/config-gating
        flags = {
            "debug": False,
            "advancements_disabled": False,
            "testing_mode": False,
        }

        # first, resolve exported flags by OR
        for modid, plugin in self.mod_plugins.items():
            for flag, value in plugin.flags.items():
                if flag in flags:
                    resolved = flags[flag] = flags[flag] or value
                    logger.debug(
                        f"{modid} exports flag {flag}={value}, resolving to {resolved}"
                    )
                else:
                    flags[flag] = value
                    logger.debug(f"{modid} exports flag {flag}={value}")

        # then, apply local overrides
        flags |= self.props.flags

        # any missing flags will default to True

        logger.debug(f"Resolved flags: {flags}")
        return flags

    @overload
    def _hook_caller(self, spec: Callable[_P, None]) -> _NoCallTypedHookCaller[_P]: ...

    @overload
    def _hook_caller(
        self, spec: Callable[_P, _R | None]
    ) -> TypedHookCaller[_P, _R]: ...

    def _hook_caller(self, spec: Callable[_P, _R | None]) -> TypedHookCaller[_P, _R]:
        caller = self.inner.hook.__dict__[spec.__name__]
        return TypedHookCaller(None, caller)


def flatten(values: list[list[_T] | _T] | None) -> Iterator[_T]:
    for value in values or []:
        if isinstance(value, list):
            yield from value
        else:
            yield value


def import_package(package: Package) -> ModuleType:
    match package:
        case ModuleType():
            return package
        case str():
            return importlib.import_module(package)



# pyright: reportUnknownMemberType=false, reportUnknownArgumentType=false
# pyright: reportUnknownLambdaType=false, reportUnknownVariableType=false

import logging
import shutil
from pathlib import Path
from typing import Any, Mapping, Sequence

import langcodes
from _hexdoc_favicons import Favicons
from jinja2 import (
    BaseLoader,
    ChoiceLoader,
    Environment,
    PrefixLoader,
    StrictUndefined,
    Template,
    TemplateNotFound,
)
from jinja2.sandbox import SandboxedEnvironment

from hexdoc.core import MinecraftVersion, Properties, ResourceLocation
from hexdoc.core.properties import JINJA_NAMESPACE_ALIASES
from hexdoc.data.sitemap import MARKER_NAME, LatestSitemapMarker, VersionedSitemapMarker
from hexdoc.minecraft import I18n
from hexdoc.plugin import ModPluginWithBook, PluginManager
from hexdoc.utils import ContextSource, write_to_path

from .extensions import DefaultMacroExtension, IncludeRawExtension
from .filters import (
    hexdoc_item,
    hexdoc_localize,
    hexdoc_smart_var,
    hexdoc_texture,
    hexdoc_wrap,
)

logger = logging.getLogger(__name__)


class HexdocTemplateLoader(BaseLoader):
    def __init__(
        self,
        included: dict[str, BaseLoader],
        extra: dict[str, BaseLoader],
        props_file: Path,
    ):
        self.inner = ChoiceLoader(
            [
                PrefixLoader(included, ":"),
                *included.values(),
            ]
        )
        self.extra = extra
        self.props_file = props_file

    def get_source(self, environment: Environment, template: str):
        for alias, replacement in JINJA_NAMESPACE_ALIASES.items():
            if template.startswith(f"{alias}:"):
                logger.debug(
                    f"Replacing {alias} with {replacement} for template {template}"
                )
                template = replacement + template.removeprefix(alias)
                break

        try:
            return self.inner.get_source(environment, template)
        except TemplateNotFound as e:
            for modid, loader in self.extra.items():
                try:
                    loader.get_source(environment, template)
                except TemplateNotFound:
                    continue
                e.add_note(
                    f'  note: try adding "{modid}" to props.template.include '
                    f"in {self.props_file.as_posix()}"
                )
            raise


def create_jinja_env(pm: PluginManager, include: Sequence[str], props_file: Path):
    included, extra = pm.load_jinja_templates(include)

    env = create_jinja_env_with_loader(
        HexdocTemplateLoader(
            included=included,
            extra=extra,
            props_file=props_file,
        )
    )

    return pm.update_jinja_env(include, env)


def create_jinja_env_with_loader(loader: BaseLoader):
    env = SandboxedEnvironment(
        loader=loader,
        undefined=StrictUndefined,
        lstrip_blocks=True,
        trim_blocks=True,
        autoescape=True,
        extensions=[
            IncludeRawExtension,
            DefaultMacroExtension,
        ],
    )

    env.filters |= {  # pyright: ignore[reportAttributeAccessIssue]
        "hexdoc_wrap": hexdoc_wrap,
        "hexdoc_localize": hexdoc_localize,
        "hexdoc_texture": hexdoc_texture,
        "hexdoc_item": hexdoc_item,
        "hexdoc_smart_var": hexdoc_smart_var,
    }

    return env


def get_templates(
    *,
    props: Properties,
    pm: PluginManager,
    book: Any,
    context: ContextSource,
    env: Environment,
) -> dict[Path, tuple[Template, Mapping[str, Any]]]:
    assert props.template is not None

    if props.template.override_default_render:
        template_names = {k: (v, {}) for k, v in props.template.render.items()}
    else:
        template_names = pm.default_rendered_templates(
            modids=props.template.render_from,
            book=book,
            context=context,
        )

    template_names |= {k: (v, {}) for k, v in props.template.extend_render.items()}

    return {
        Path(path): (env.get_template(template_name), args)
        for path, (template_name, args) in template_names.items()
    }


def render_book(
    *,
    props: Properties,
    pm: PluginManager,
    plugin: ModPluginWithBook,
    lang: str,
    book_id: ResourceLocation,
    i18n: I18n,
    macros: dict[str, str],
    env: Environment,
    templates: dict[Path, tuple[Template, Mapping[str, Any]]],
    output_dir: Path,
    site_path: Path,
    version: str,
    versioned: bool,
    template_args: dict[str, Any],
):
    if not props.template:
        raise ValueError("Expected a value for props.template, got None")

    output_dir /= site_path
    page_url = props.env.github_pages_url.joinpath(*site_path.parts)

    # eg. `/v/latest/main/en_us` -> `../../../..`, `/` -> `.`
    relative_site_url = "/".join(".." for _ in site_path.parts) or "."

    logger.info(f"Rendering {output_dir}.")

    output_dir.mkdir(parents=True, exist_ok=True)

    if icon := props.template.icon:
        icon_href = icon.name

        shutil.copy(icon, output_dir)

        with Favicons(icon, output_dir, base_url="") as favicons:
            favicons.sgenerate()
            favicons_html = favicons.html()
            favicons_formats = favicons.formats()
    else:
        icon_href = None
        favicons_html = []
        favicons_formats = []

    lang_name = i18n.localize_lang()

    minecraft_version = MinecraftVersion.get()

    template_args |= {
        "props": props,
        "i18n": i18n,
        "site_url": str(props.env.github_pages_url),
        "relative_site_url": relative_site_url,
        "page_url": str(page_url),
        "source_url": str(props.env.source_url),
        "version": version,
        "lang": lang,
        "lang_name": lang_name,
        "lang_html": langcodes.standardize_tag(lang),
        "is_bleeding_edge": version.startswith("latest"),
        "favicons_html": favicons_html,
        "favicons_formats": favicons_formats,
        "icon_href": icon_href,
        "safari_pinned_tab_href": "https://raw.githubusercontent.com/hexdoc-dev/hexdoc/main/media/safari-pinned-tab.svg",
        "safari_pinned_tab_color": "#332233",
        "minecraft_version": minecraft_version or "???",
        "full_version": plugin.full_version,
        "navbar": {  # default navbar links (ignored if set in props)
            "center": [
                {"text": "GitHub", "href": {"variable": "source_url"}},
            ],
        },
        "_": lambda key: hexdoc_localize(  # i18n helper
            key,
            do_format=False,
            props=props,
            book_id=book_id,
            i18n=i18n,
            macros=macros,
            pm=pm,
        ),
        "_f": lambda key: hexdoc_localize(  # i18n helper with patchi formatting
            key,
            do_format=True,
            props=props,
            book_id=book_id,
            i18n=i18n,
            macros=macros,
            pm=pm,
        ),
        **props.template.args,
    }
    pm.update_template_args(template_args)

    pm.pre_render_book(template_args, output_dir, props.template.include)

    for filename, (template, extra_args) in templates.items():
        file = template.render(template_args | dict(extra_args))
        stripped_file = strip_empty_lines(file)
        write_to_path(output_dir / filename, stripped_file)

    if props.template.static_dir:
        shutil.copytree(props.template.static_dir, output_dir, dirs_exist_ok=True)

    # redirect file for this book
    if props.template.redirect:
        _, redirect_template_name = props.template.redirect
        redirect_template = env.get_template(redirect_template_name)
        redirect_contents = strip_empty_lines(redirect_template.render(template_args))
    else:
        redirect_contents = ""

    # marker file for updating the sitemap later
    # we use this because matrix doesn't have outputs
    # this feels scuffed but it does work
    marker_path = "/" + "/".join(site_path.parts)
    is_default_lang = lang == props.default_lang

    if versioned:
        marker = VersionedSitemapMarker(
            version=version,
            lang=lang,
            lang_name=lang_name,
            path=marker_path,
            is_default_lang=is_default_lang,
            full_version=plugin.full_version,
            minecraft_version=minecraft_version,
            redirect_contents=redirect_contents,
            mod_version=plugin.mod_version,
            plugin_version=plugin.plugin_version,
        )
    else:
        marker = LatestSitemapMarker(
            version=version,
            lang=lang,
            lang_name=lang_name,
            path=marker_path,
            is_default_lang=is_default_lang,
            full_version=plugin.full_version,
            minecraft_version=minecraft_version,
            redirect_contents=redirect_contents,
            branch=plugin.branch,
            is_default_branch=plugin.branch == props.default_branch,
        )

    (output_dir / MARKER_NAME).write_text(marker.model_dump_json(), "utf-8")

    pm.post_render_book(template_args, output_dir, props.template.include)


def strip_empty_lines(text: str) -> str:
    return "\n".join(s for s in text.splitlines() if s.strip())



# pyright: reportInvalidTypeVarUse=information

from __future__ import annotations

import logging
import subprocess
from collections.abc import Iterator
from contextlib import ExitStack
from pathlib import Path
from textwrap import dedent
from typing import Any, Callable, Literal, Self, Sequence, TypeVar, overload

from pydantic import SkipValidation
from pydantic.dataclasses import dataclass

from hexdoc.model import DEFAULT_CONFIG, HexdocModel
from hexdoc.plugin import PluginManager
from hexdoc.utils import (
    TRACE,
    JSONDict,
    ValidationContext,
    decode_json_dict,
    must_yield_something,
    strip_suffixes,
    write_to_path,
)
from hexdoc.utils.cd import relative_path_root
from hexdoc.utils.types import PydanticOrderedSet

from .properties import Properties
from .resource import ResourceLocation, ResourceType
from .resource_dir import PathResourceDir

logger = logging.getLogger(__name__)

METADATA_SUFFIX = ".hexdoc.json"


_T = TypeVar("_T")
_T_Model = TypeVar("_T_Model", bound=HexdocModel)

ExportFn = Callable[[_T, _T | None], str]

BookFolder = Literal["categories", "entries", "templates"]


@dataclass(config=DEFAULT_CONFIG | {"arbitrary_types_allowed": True}, kw_only=True)
class ModResourceLoader(ValidationContext):
    props: Properties
    export_dir: Path | None
    resource_dirs: Sequence[PathResourceDir]
    _stack: SkipValidation[ExitStack]

    @classmethod
    def clean_and_load_all(
        cls,
        props: Properties,
        pm: PluginManager,
        *,
        export: bool = False,
    ):
        # clear the export dir so we start with a clean slate
        if props.export_dir and export:
            subprocess.run(
                ["git", "clean", "-fdX", props.export_dir],
                cwd=props.props_dir,
            )

            write_to_path(
                props.export_dir / "__init__.py",
                dedent(
                    """\
                    # This directory is auto-generated by hexdoc.
                    # Do not edit or commit these files.
                    """
                ),
            )

        return cls.load_all(
            props,
            pm,
            export=export,
        )

    @classmethod
    def load_all(
        cls,
        props: Properties,
        pm: PluginManager,
        *,
        export: bool = False,
    ) -> Self:
        export_dir = props.export_dir if export else None
        stack = ExitStack()

        with relative_path_root(Path()):
            resource_dirs = [
                path_resource_dir
                for resource_dir in props.resource_dirs
                for path_resource_dir in stack.enter_context(resource_dir.load(pm))
            ]

        return cls(
            props=props,
            export_dir=export_dir,
            resource_dirs=resource_dirs,
            _stack=stack,
        )

    def __enter__(self):
        return self

    def __exit__(self, *exc_details: Any):
        return self._stack.__exit__(*exc_details)

    def close(self):
        self._stack.close()

    def _map_own_assets(self, folder: str, *, root: str | Path):
        return {
            id: path.resolve().relative_to(root)
            for _, id, path in self.find_resources(
                "assets",
                namespace=self.props.modid,
                folder="",
                glob=f"{folder}/**/*.*",
                allow_missing=True,
            )
        }

    @property
    def should_export(self):
        return self.export_dir is not None

    def load_metadata(
        self,
        *,
        name_pattern: str = "{modid}",
        model_type: type[_T_Model],
        allow_missing: bool = False,
    ) -> dict[str, _T_Model]:
        """eg. `"{modid}.patterns"`"""
        metadata = dict[str, _T_Model]()

        # TODO: refactor
        cached_metadata = self.props.cache_dir / (
            name_pattern.format(modid=self.props.modid) + METADATA_SUFFIX
        )
        if cached_metadata.is_file():
            metadata[self.props.modid] = model_type.model_validate_json(
                cached_metadata.read_bytes()
            )

        for resource_dir in self.resource_dirs:
            # skip if the resource dir has no metadata set, because we're only loading
            # this for external mods (TODO: this feels flawed)
            modid = resource_dir.modid
            if modid is None or modid in metadata:
                continue

            try:
                _, metadata[modid] = self.load_resource(
                    Path(name_pattern.format(modid=modid) + METADATA_SUFFIX),
                    decode=model_type.model_validate_json,
                    export=False,
                )
            except FileNotFoundError:
                if allow_missing:
                    continue
                raise

        return metadata

    @must_yield_something
    def load_book_assets(
        self,
        parent_book_id: ResourceLocation,
        folder: BookFolder,
        use_resource_pack: bool,
        lang: str | None = None,
    ) -> Iterator[tuple[PathResourceDir, ResourceLocation, JSONDict]]:
        if self.props.book_id is None:
            raise TypeError("Can't load book assets because props.book_id is None")

        if lang is None:
            lang = self.props.default_lang

        # use ordered set to be deterministic but avoid duplicate ids
        books_to_check = PydanticOrderedSet[ResourceLocation].collect(
            parent_book_id,
            self.props.book_id,
            *self.props.extra_books,
        )

        for book_id in books_to_check:
            yield from self.load_resources(
                type="assets" if use_resource_pack else "data",
                folder=Path("patchouli_books") / book_id.path / lang / folder,
                namespace=book_id.namespace,
                allow_missing=True,
            )

    @overload
    def load_resource(
        self,
        type: ResourceType,
        folder: str | Path,
        id: ResourceLocation,
        *,
        decode: Callable[[str], _T] = decode_json_dict,
        export: ExportFn[_T] | Literal[False] | None = None,
    ) -> tuple[PathResourceDir, _T]: ...

    @overload
    def load_resource(
        self,
        path: Path,
        /,
        *,
        decode: Callable[[str], _T] = decode_json_dict,
        export: ExportFn[_T] | Literal[False] | None = None,
    ) -> tuple[PathResourceDir, _T]: ...

    def load_resource(
        self,
        *args: Any,
        decode: Callable[[str], _T] = decode_json_dict,
        export: ExportFn[_T] | Literal[False] | None = None,
        **kwargs: Any,
    ) -> tuple[PathResourceDir, _T]:
        """Find the first file with this resource location in `resource_dirs`.

        If no file extension is provided, `.json` is assumed.

        Raises FileNotFoundError if the file does not exist.
        """

        resource_dir, path = self.find_resource(*args, **kwargs)
        return resource_dir, self._load_path(
            resource_dir,
            path,
            decode=decode,
            export=export,
        )

    @overload
    def find_resource(
        self,
        type: ResourceType,
        folder: str | Path,
        id: ResourceLocation,
    ) -> tuple[PathResourceDir, Path]: ...

    @overload
    def find_resource(
        self,
        path: Path,
        /,
    ) -> tuple[PathResourceDir, Path]: ...

    def find_resource(
        self,
        type: ResourceType | Path,
        folder: str | Path | None = None,
        id: ResourceLocation | None = None,
    ) -> tuple[PathResourceDir, Path]:
        """Find the first file with this resource location in `resource_dirs`.

        If no file extension is provided, `.json` / `.json5` is assumed.

        Raises FileNotFoundError if the file does not exist.
        """

        if isinstance(type, Path):
            path_stub = type
        else:
            assert folder is not None and id is not None
            path_stub = id.file_path_stub(type, folder)

        # check by descending priority, return the first that exists
        for resource_dir in self.resource_dirs:
            path = resource_dir.path / path_stub
            if path.is_file():
                return resource_dir, path
            if path.suffix == ".json":
                path = path.with_suffix(".json5")
                if path.is_file():
                    return resource_dir, path

        raise FileNotFoundError(f"Path {path_stub} not found in any resource dir")

    @overload
    def load_resources(
        self,
        type: ResourceType,
        *,
        namespace: str,
        folder: str | Path,
        glob: str | list[str] = "**/*",
        allow_missing: bool = False,
        internal_only: bool = False,
        decode: Callable[[str], _T] = decode_json_dict,
        export: ExportFn[_T] | Literal[False] | None = None,
    ) -> Iterator[tuple[PathResourceDir, ResourceLocation, _T]]: ...

    @overload
    def load_resources(
        self,
        type: ResourceType,
        *,
        folder: str | Path,
        id: ResourceLocation,
        allow_missing: bool = False,
        internal_only: bool = False,
        decode: Callable[[str], _T] = decode_json_dict,
        export: ExportFn[_T] | Literal[False] | None = None,
    ) -> Iterator[tuple[PathResourceDir, ResourceLocation, _T]]: ...

    def load_resources(
        self,
        type: ResourceType,
        *,
        decode: Callable[[str], _T] = decode_json_dict,
        export: ExportFn[_T] | Literal[False] | None = None,
        **kwargs: Any,
    ) -> Iterator[tuple[PathResourceDir, ResourceLocation, _T]]:
        """Like `find_resources`, but also loads the file contents and reexports it."""
        for resource_dir, value_id, path in self.find_resources(type, **kwargs):
            value = self._load_path(
                resource_dir,
                path,
                decode=decode,
                export=export,
            )
            yield resource_dir, value_id, value

    @overload
    def find_resources(
        self,
        type: ResourceType,
        *,
        namespace: str,
        folder: str | Path,
        glob: str | list[str] = "**/*",
        allow_missing: bool = False,
        internal_only: bool = False,
    ) -> Iterator[tuple[PathResourceDir, ResourceLocation, Path]]: ...

    @overload
    def find_resources(
        self,
        type: ResourceType,
        *,
        folder: str | Path,
        id: ResourceLocation,
        allow_missing: bool = False,
        internal_only: bool = False,
    ) -> Iterator[tuple[PathResourceDir, ResourceLocation, Path]]: ...

    def find_resources(
        self,
        type: ResourceType,
        *,
        folder: str | Path,
        id: ResourceLocation | None = None,
        namespace: str | None = None,
        glob: str | list[str] = "**/*",
        allow_missing: bool = False,
        internal_only: bool = False,
    ) -> Iterator[tuple[PathResourceDir, ResourceLocation, Path]]:
        """Search for a glob under a given resource location in all of `resource_dirs`.

        Files are returned from lowest to highest priority in the load order, ie. later
        files should overwrite earlier ones.

        If no file extension is provided for glob, `.json` is assumed.

        Raises FileNotFoundError if no files were found in any resource dir.

        For example:
        ```py
        props.find_resources(
            "assets",
            "lang/subdir",
            namespace="*",
            glob="*.flatten.json5",
        )

        # [(hexcasting:en_us, .../resources/assets/hexcasting/lang/subdir/en_us.json)]
        ```
        """

        if id is not None:
            namespace = id.namespace
            glob = id.path

        # eg. assets/*/lang/subdir
        if namespace is not None:
            base_path_stub = Path(type) / namespace / folder
        else:
            raise RuntimeError(
                "No overload matches the specified arguments (expected id or namespace)"
            )

        # glob for json files if not provided
        globs = [glob] if isinstance(glob, str) else glob
        for i in range(len(globs)):
            if not Path(globs[i]).suffix:
                globs.append(globs[i] + ".json5")
                globs[i] += ".json"

        # find all files matching the resloc
        found_any = False
        for resource_dir in reversed(self.resource_dirs):
            if internal_only and not resource_dir.internal:
                continue

            # eg. .../resources/assets/*/lang/subdir
            for base_path in resource_dir.path.glob(base_path_stub.as_posix()):
                for glob_ in globs:
                    # eg. .../resources/assets/hexcasting/lang/subdir/*.flatten.json5
                    for path in base_path.glob(glob_):
                        # only strip json/json5, not eg. png
                        id_path = path.relative_to(base_path)
                        if "json" in path.name:
                            id_path = strip_suffixes(id_path)

                        id = ResourceLocation(
                            # eg. ["assets", "hexcasting", "lang", ...][1]
                            namespace=path.relative_to(resource_dir.path).parts[1],
                            path=id_path.as_posix(),
                        )

                        if path.is_file():
                            found_any = True
                            yield resource_dir, id, path

        # if we never yielded any files, raise an error
        if not allow_missing and not found_any:
            raise FileNotFoundError(
                f"No files found under {base_path_stub / repr(globs)} in any resource dir"
            )

    def _load_path(
        self,
        resource_dir: PathResourceDir,
        path: Path,
        *,
        decode: Callable[[str], _T] = decode_json_dict,
        export: ExportFn[_T] | Literal[False] | None = None,
    ) -> _T:
        if not path.is_file():
            raise FileNotFoundError(path)

        logger.debug(f"Loading {path}")

        data = path.read_text("utf-8")
        value = decode(data)

        if resource_dir.reexport and export is not False:
            self.export(
                path.relative_to(resource_dir.path),
                data,
                value,
                decode=decode,
                export=export,
            )

        return value

    @overload
    def export(self, /, path: Path, data: str, *, cache: bool = False) -> None: ...

    @overload
    def export(
        self,
        /,
        path: Path,
        data: str,
        value: _T,
        *,
        decode: Callable[[str], _T] = decode_json_dict,
        export: ExportFn[_T] | None = None,
        cache: bool = False,
    ) -> None: ...

    def export(
        self,
        path: Path,
        data: str,
        value: _T = None,
        *,
        decode: Callable[[str], _T] = decode_json_dict,
        export: ExportFn[_T] | None = None,
        cache: bool = False,
    ) -> None:
        if not self.export_dir:
            return
        out_path = self.export_dir / path

        logger.log(TRACE, f"Exporting {path} to {out_path}")
        if export is None:
            out_data = data
        else:
            try:
                old_value = decode(out_path.read_text("utf-8"))
            except FileNotFoundError:
                old_value = None

            out_data = export(value, old_value)

        write_to_path(out_path, out_data)

        if cache:
            write_to_path(self.props.cache_dir / path, out_data)

    def export_raw(self, path: Path, data: bytes):
        if not self.export_dir:
            return
        out_path = self.export_dir / path

        logger.log(TRACE, f"Exporting {path} to {out_path}")
        write_to_path(out_path, data)

    def __repr__(self):
        return f"{self.__class__.__name__}(...)"



from __future__ import annotations

from typing import Self

from typing_extensions import override

from hexdoc.core import ResourceLocation
from hexdoc.model import HexdocModel

from .base_model import BaseMinecraftModel


class ItemModel(BaseMinecraftModel):
    """Represents a Minecraft item model.

    https://minecraft.wiki/w/Tutorials/Models#Item_models
    """

    overrides: list[ItemModelOverride] | None = None
    """Determines cases in which a different model should be used based on item tags.

    All cases are evaluated in order from top to bottom and last predicate that matches
    overrides. However, overrides are ignored if it has been already overridden once,
    for example this avoids recursion on overriding to the same model.
    """

    @override
    def apply_parent(self, parent: Self):
        super().apply_parent(parent)


class ItemModelOverride(HexdocModel):
    """An item model override case.

    https://minecraft.wiki/w/Tutorials/Models#Item_models
    """

    model: ResourceLocation
    """The path to the model to use if the case is met."""
    predicate: dict[ResourceLocation, float]
    """Item predicates that must be true for this model to be used."""



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from importlib.resources import Package
from pathlib import Path
from typing import TYPE_CHECKING, Any, Mapping

from jinja2.sandbox import SandboxedEnvironment
from typing_extensions import override
from yarl import URL

from hexdoc.utils import ContextSource

from .types import HookReturn

if TYPE_CHECKING:
    from hexdoc.cli.app import LoadedBookInfo
    from hexdoc.core import ModResourceLoader, Properties
    from hexdoc.minecraft.assets import HexdocAssetLoader

DefaultRenderedTemplates = Mapping[
    str | Path,
    str | tuple[str, Mapping[str, Any]],
]


@dataclass(kw_only=True)
class ModPlugin(ABC):
    """Hexdoc plugin hooks that are tied to a specific Minecraft mod.

    If you want to render a web book, subclass `ModPluginWithBook` instead.

    Abstract methods are required. All other methods can optionally be implemented to
    override or add functionality to hexdoc.

    Non-mod-specific hooks are implemented with normal Pluggy hooks instead.
    """

    branch: str
    props: Properties | None = None

    # required hooks

    @property
    @abstractmethod
    def modid(self) -> str:
        """The modid of the Minecraft mod version that this plugin represents.

        For example: `hexcasting`
        """

    @property
    @abstractmethod
    def full_version(self) -> str:
        """The full PyPI version of this plugin.

        This should generally return `your_plugin.__gradle_version__.FULL_VERSION`.

        For example: `0.11.1.1.0rc7.dev20`
        """

    @property
    @abstractmethod
    def plugin_version(self) -> str:
        """The hexdoc-specific component of this plugin's version number.

        This should generally return `your_plugin.__version__.PY_VERSION`.

        For example: `1.0.dev20`
        """

    # optional hooks

    @property
    def compat_minecraft_version(self) -> str | None:
        """The version of Minecraft supported by the mod that this plugin represents.

        If no plugins implement this, models and validation for all Minecraft versions
        may be used. Currently, if two or more plugins provide different values, an
        error will be raised.

        This should generally return `your_plugin.__gradle_version__.MINECRAFT_VERSION`.

        For example: `1.20.1`
        """
        return None

    @property
    def mod_version(self) -> str | None:
        """The Minecraft mod version that this plugin represents.

        This should generally return `your_plugin.__gradle_version__.GRADLE_VERSION`.

        For example: `0.11.1-7`
        """
        return None

    def resource_dirs(self) -> HookReturn[Package]:
        """The module(s) that contain your plugin's Minecraft resources to be rendered.

        For example: `your_plugin._export.generated`
        """
        return []

    def jinja_template_root(self) -> HookReturn[tuple[Package, str]] | None:
        """The module that contains the folder with your plugin's Jinja templates, and
        the name of that folder.

        For example: `your_plugin, "_templates"`
        """
        return None

    def default_rendered_templates(self) -> DefaultRenderedTemplates:
        """Extra templates to be rendered by default when your plugin is active.

        The key is the output path, and the value is the template to import and render.
        It may also be a tuple where the first item is the template and the second is
        a dict to be merged with the arguments for that template.

        This hook is not called if `props.template.render` is set, since that option
        overrides all default templates.
        """
        return {}

    def default_rendered_templates_v2(
        self,
        book: Any,
        context: ContextSource,
    ) -> DefaultRenderedTemplates:
        """Like `default_rendered_templates`, but gets access to the book and context.

        This is useful for dynamically generating multi-file output structures.
        """
        return {}

    def update_jinja_env(self, env: SandboxedEnvironment) -> None:
        """Modify the Jinja environment/configuration.

        This is called after hexdoc is done setting up the Jinja environment but before
        rendering the book.
        """

    def pre_render_site(
        self,
        props: Properties,
        books: list[LoadedBookInfo],
        env: SandboxedEnvironment,
        output_dir: Path,
    ) -> None:
        """Called once, after hexdoc is done setting up the Jinja environment but before
        rendering the book."""

    def post_render_site(
        self,
        props: Properties,
        books: list[LoadedBookInfo],
        env: SandboxedEnvironment,
        output_dir: Path,
    ) -> None:
        """Called once, at the end of `hexdoc build`."""

    def pre_render_book(self, template_args: dict[str, Any], output_dir: Path) -> None:
        """Called once per language, just before rendering the book for that
        language."""

    def post_render_book(self, template_args: dict[str, Any], output_dir: Path) -> None:
        """Called once per language, after all book files for that language are
        rendered."""

    # utils

    def site_path(self, versioned: bool):
        if versioned:
            return self.versioned_site_path
        return self.latest_site_path

    @property
    def site_root(self) -> Path:
        """Base path for all rendered web pages.

        For example:
        * URL: `https://gamma-delta.github.io/HexMod/v/0.11.1-7/1.0.dev20/en_us`
        * value: `v`
        """
        return Path("v")

    @property
    def versioned_site_path(self) -> Path:
        """Base path for the web pages for the current version.

        For example:
        * URL: `https://hexdoc.hexxy.media/book/v/1!0.1.0.dev0` (decoded)
        * value: `book/v/1!0.1.0.dev0`
        """
        return self.site_root / self.full_version

    @property
    def latest_site_path(self) -> Path:
        """Base path for the latest web pages for a given branch.

        For example:
        * URL: `https://gamma-delta.github.io/HexMod/v/latest/main/en_us`
        * value: `v/latest/main`
        """
        return self.site_root / "latest" / self.branch

    def asset_loader(
        self,
        loader: ModResourceLoader,
        *,
        site_url: URL,
        asset_url: URL,
        render_dir: Path,
    ) -> HexdocAssetLoader:
        # unfortunately, this is necessary to avoid some *real* ugly circular imports
        from hexdoc.minecraft.assets import HexdocAssetLoader

        return HexdocAssetLoader(
            loader=loader,
            site_url=site_url,
            asset_url=asset_url,
            render_dir=render_dir,
        )

    @property
    def flags(self) -> dict[str, bool]:
        """Set default values for Patchouli config flags.

        Unlike the table in `hexdoc.toml`, this is available for use in dependents.
        """
        return {}


class VersionedModPlugin(ModPlugin):
    """Like `ModPlugin`, but the versioned site path uses the plugin and mod version."""

    @property
    @abstractmethod
    @override
    def mod_version(self) -> str: ...

    @property
    @override
    def versioned_site_path(self) -> Path:
        """Base path for the web pages for the current version.

        For example:
        * URL: `https://gamma-delta.github.io/HexMod/v/0.11.1-7/1.0.dev20/en_us`
        * value: `v/0.11.1-7/1.0.dev20`
        """
        return self.site_root / self.mod_version / self.plugin_version


class ModPluginWithBook(VersionedModPlugin):
    """Like `ModPlugin`, but with extra hooks to support rendering a web book."""

    @abstractmethod
    @override
    def resource_dirs(self) -> HookReturn[Package]: ...

    def site_book_path(self, lang: str, versioned: bool) -> Path:
        if versioned:
            return self.versioned_site_book_path(lang)
        return self.latest_site_book_path(lang)

    def versioned_site_book_path(self, lang: str) -> Path:
        """Base path for the rendered web book for the current version.

        For example:
        * URL: `https://gamma-delta.github.io/HexMod/v/0.11.1-7/1.0.dev20/en_us`
        * value: `v/0.11.1-7/1.0.dev20/en_us`
        """
        return self.versioned_site_path / lang

    def latest_site_book_path(self, lang: str) -> Path:
        """Base path for the latest rendered web book for a given branch.

        For example:
        * URL: `https://gamma-delta.github.io/HexMod/v/latest/main/en_us`
        * value: `v/latest/main/en_us`
        """
        return self.latest_site_path / lang



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import (
    Annotated,
    Any,
    Iterable,
    Literal,
    Self,
    TypeVar,
)

from pydantic import Field, SerializeAsAny
from typing_extensions import override
from yarl import URL

from hexdoc.core import ResourceLocation
from hexdoc.model import (
    InlineModel,
    ValidationContextModel,
)
from hexdoc.utils import ContextSource, PydanticURL

from .constants import MISSING_TEXTURE_URL

logger = logging.getLogger(__name__)


class BaseTexture(InlineModel, ABC):
    @classmethod
    @abstractmethod
    def from_url(cls, url: URL, pixelated: bool) -> Self: ...

    @override
    @classmethod
    def load_id(cls, id: ResourceLocation, context: ContextSource):
        texture_ctx = TextureContext.of(context)
        return cls.lookup(
            id,
            lookups=texture_ctx.textures,
            allowed_missing=texture_ctx.allowed_missing_textures,
        )

    @classmethod
    def lookup(
        cls,
        id: ResourceLocation,
        lookups: TextureLookups[Any],
        allowed_missing: Iterable[ResourceLocation] | Literal["*"],
    ) -> Self:
        """Returns the texture from the lookup table if it exists, or the "missing
        texture" texture if it's in `props.texture.missing`, or raises `ValueError`.

        This is called frequently and does not load any files.
        """
        textures = cls.get_lookup(lookups)
        if id in textures:
            return textures[id]

        # TODO: this logic is duplicated in load_assets.py :/
        if allowed_missing == "*" or any(
            id.match(pattern) for pattern in allowed_missing
        ):
            logger.warning(f"No {cls.__name__} for {id}, using default missing texture")
            return cls.from_url(MISSING_TEXTURE_URL, pixelated=True)

        raise ValueError(f"No {cls.__name__} for {id}")

    @classmethod
    def get_lookup(cls, lookups: TextureLookups[Any]) -> TextureLookup[Self]:
        return lookups[cls.__name__]

    def insert_texture(self, lookups: TextureLookups[Any], id: ResourceLocation):
        textures = self.get_lookup(lookups)
        textures[id] = self


class PNGTexture(BaseTexture):
    url: PydanticURL | None
    pixelated: bool

    @classmethod
    def from_url(cls, url: URL, pixelated: bool) -> Self:
        return cls(url=url, pixelated=pixelated)


_T_BaseTexture = TypeVar("_T_BaseTexture", bound=BaseTexture)

TextureLookup = dict[ResourceLocation, SerializeAsAny[_T_BaseTexture]]
"""dict[id, texture]"""

TextureLookups = defaultdict[
    str,
    Annotated[TextureLookup[_T_BaseTexture], Field(default_factory=dict)],
]
"""dict[type(texture).__name__, TextureLookup]"""


class TextureContext(ValidationContextModel):
    textures: TextureLookups[Any]
    allowed_missing_textures: set[ResourceLocation] | Literal["*"]



from __future__ import annotations

from typing import Self

from typing_extensions import override

from hexdoc.core import ModResourceLoader, ResourceLocation

from .base_model import BaseMinecraftModel


class BlockModel(BaseMinecraftModel):
    """Represents a Minecraft block model.

    https://minecraft.wiki/w/Tutorials/Models#Block_models
    """

    ambientocclusion: bool = True
    """Whether to use ambient occlusion or not.

    Note: only works on parent file.
    """
    render_type: ResourceLocation | None = None
    """Sets the rendering type for this model.

    https://docs.minecraftforge.net/en/latest/rendering/modelextensions/rendertypes/
    """

    @override
    def apply_parent(self, parent: Self):
        super().apply_parent(parent)
        self.ambientocclusion = parent.ambientocclusion
        self.render_type = self.render_type or parent.render_type

    def load_parents_and_apply(self, loader: ModResourceLoader):
        while self.parent:
            _, parent = loader.load_resource(
                "assets",
                "models",
                self.parent,
                decode=self.model_validate_json,
            )
            self.apply_parent(parent)

    def resolve_texture_variables(self):
        textures = dict[str, ResourceLocation]()
        for name, value in self.textures.items():
            while not isinstance(value, ResourceLocation):
                value = value.lstrip("#")
                if value == name:
                    raise ValueError(f"Cyclic texture variable detected: {name}")
                value = self.textures[value]
            textures[name] = value
        return textures



import string
from typing import Any

from pydantic import field_validator, model_validator
from pydantic.dataclasses import dataclass

from hexdoc.utils.json_schema import inherited, json_schema_extra_config, type_str

from .base import DEFAULT_CONFIG


@dataclass(
    frozen=True,
    config=DEFAULT_CONFIG
    | json_schema_extra_config(
        type_str,
        inherited,
        pattern=r"(#|0x)?([0-9a-fA-F]{6}|[0-9a-fA-F]{3})",
    ),
)
class Color:
    """Represents a hexadecimal color.

    Inputs are coerced to lowercase `rrggbb`. Raises ValueError on invalid input.

    Valid formats, all of which would be converted to `0099ff`:
    - `"#0099FF"`
    - `"#0099ff"`
    - `"#09F"`
    - `"#09f"`
    - `"0099FF"`
    - `"0099ff"`
    - `"09F"`
    - `"09f"`
    - `0x0099ff`
    """

    value: str

    @model_validator(mode="before")
    def _pre_root(cls, value: Any):
        if isinstance(value, (str, int)):
            return {"value": value}
        return value

    @field_validator("value", mode="before")
    def _check_value(cls, value: Any) -> str:
        # type check
        match value:
            case str():
                value = value.removeprefix("#").lower()
            case int():
                # int to hex string
                value = f"{value:0>6x}"
            case _:
                raise TypeError(f"Expected str or int, got {type(value)}")

        # 012 -> 001122
        if len(value) == 3:
            value = "".join(c + c for c in value)

        # length and character check
        if len(value) != 6 or any(c not in string.hexdigits for c in value):
            raise ValueError(f"invalid color code: {value}")

        return value



# pyright: reportPrivateUsage=false

from importlib.resources import Package
from typing import Any, Mapping

from hexdoc.core import ModResourceLoader, ResourceLocation

# from hexdoc.patchouli import BookContext
from hexdoc.plugin import (
    BookPlugin,
    BookPluginImpl,
    HookReturn,
    LoadTaggedUnionsImpl,
    hookimpl,
)
from hexdoc.utils import ContextSource, JSONDict, cast_context

from hexdoc_modonomicon.book import Modonomicon


class ModonomiconPlugin(LoadTaggedUnionsImpl, BookPluginImpl):
    @staticmethod
    @hookimpl
    def hexdoc_book_plugin() -> BookPlugin[Any]:
        return ModonomiconBookPlugin()

    @staticmethod
    @hookimpl
    def hexdoc_load_tagged_unions() -> HookReturn[Package]:
        return []


class ModonomiconBookPlugin(BookPlugin[Modonomicon]):
    @property
    def modid(self):
        return "modonomicon"

    def load_book_data(
        self,
        book_id: ResourceLocation,
        loader: ModResourceLoader,
    ) -> tuple[ResourceLocation, JSONDict]:
        _, data = loader.load_resource("data", "modonomicon", book_id / "book")
        return book_id, data

    def is_i18n_enabled(self, book_data: Mapping[str, Any]):
        return True

    def validate_book(
        self,
        book_data: Mapping[str, Any],
        *,
        context: ContextSource,
    ):
        # loader = ModResourceLoader.of(context)
        # book_ctx = BookContext.of(context)

        book = Modonomicon.model_validate(book_data, context=cast_context(context))
        # book._load_categories(context, book_ctx)
        # book._load_entries(context, book_ctx, loader)

        return book



__all__ = [
    "TRACE",
    "ContextSource",
    "FieldOrProperty",
    "IProperty",
    "Inherit",
    "InheritType",
    "JSONDict",
    "JSONValue",
    "NoValue",
    "NoValueType",
    "PydanticOrderedSet",
    "PydanticURL",
    "RelativePath",
    "Sortable",
    "TOMLDict",
    "TOMLValue",
    "TryGetEnum",
    "ValidationContext",
    "add_to_context",
    "cast_context",
    "cast_or_raise",
    "clamping_validator",
    "classproperty",
    "decode_and_flatten_json_dict",
    "decode_json_dict",
    "git_root",
    "isinstance_or_raise",
    "listify",
    "load_toml_with_placeholders",
    "must_yield_something",
    "relative_path_root",
    "replace_suffixes",
    "set_contextvar",
    "setup_logging",
    "sorted_dict",
    "strip_suffixes",
    "write_to_path",
]

from .cd import RelativePath, relative_path_root
from .classproperties import classproperty
from .context import ContextSource, ValidationContext, add_to_context, cast_context
from .contextmanagers import set_contextvar
from .deserialize import (
    JSONDict,
    JSONValue,
    TOMLDict,
    TOMLValue,
    cast_or_raise,
    decode_and_flatten_json_dict,
    decode_json_dict,
    isinstance_or_raise,
    load_toml_with_placeholders,
)
from .git import git_root
from .iterators import listify, must_yield_something
from .logging import TRACE, setup_logging
from .path import replace_suffixes, strip_suffixes, write_to_path
from .singletons import Inherit, InheritType, NoValue, NoValueType
from .types import (
    FieldOrProperty,
    IProperty,
    PydanticOrderedSet,
    PydanticURL,
    Sortable,
    TryGetEnum,
    clamping_validator,
    sorted_dict,
)



from __future__ import annotations

from contextvars import ContextVar
from typing import (
    Any,
    ClassVar,
    cast,
    dataclass_transform,
)

from pydantic import BaseModel, ConfigDict, TypeAdapter, ValidationInfo, model_validator
from pydantic.functional_validators import ModelBeforeValidator
from pydantic_settings import BaseSettings, SettingsConfigDict
from yarl import URL

from hexdoc.utils import ValidationContext, set_contextvar
from hexdoc.utils.classproperties import ClassPropertyDescriptor

DEFAULT_CONFIG = ConfigDict(
    extra="forbid",
    validate_default=True,
    ignored_types=(  # pyright: ignore[reportUnknownArgumentType]
        ClassPropertyDescriptor,
        URL,
    ),
)

IGNORE_EXTRA_CONFIG = DEFAULT_CONFIG | ConfigDict(
    extra="ignore",
)

_init_context_var = ContextVar[Any]("_init_context_var", default=None)


def init_context(value: Any):
    """https://docs.pydantic.dev/latest/usage/validators/#using-validation-context-with-basemodel-initialization"""
    return set_contextvar(_init_context_var, value)


@dataclass_transform()
class HexdocModel(BaseModel):
    """Base class for all Pydantic models in hexdoc.

    Sets the default model config, and overrides __init__ to allow using the
    `init_context` context manager to set validation context for constructors.
    """

    model_config = DEFAULT_CONFIG
    """@private"""

    __hexdoc_before_validator__: ClassVar[ModelBeforeValidator | None] = None

    def __init__(__pydantic_self__, **data: Any) -> None:  # type: ignore
        __tracebackhide__ = True
        __pydantic_self__.__pydantic_validator__.validate_python(
            data,
            self_instance=__pydantic_self__,
            context=_init_context_var.get(),
        )

    __init__.__pydantic_base_init__ = True  # type: ignore

    @model_validator(mode="before")
    @classmethod
    def _call_hexdoc_before_validator(cls, value: Any, info: ValidationInfo):
        # allow json schema field in all models
        if isinstance(value, dict):
            value = cast(dict[Any, Any], value)
            value.pop("$schema", None)
        if cls.__hexdoc_before_validator__:
            return cls.__hexdoc_before_validator__(cls, value, info)
        return value


class HexdocSettings(BaseSettings):
    model_config = SettingsConfigDict(
        env_file=".env",
        extra="allow",
    )

    @classmethod
    def model_getenv(cls, defaults: Any = None):
        return cls.model_validate(defaults or {})


class ValidationContextModel(HexdocModel, ValidationContext):
    pass


HexdocTypeAdapter = TypeAdapter
"""DEPRECATED: Use `pydantic.TypeAdapter` instead."""



__all__ = [
    "HEXDOC_PROJECT_NAME",
    "BookPlugin",
    "BookPluginImpl",
    "DefaultRenderedTemplates",
    "HookReturn",
    "LoadTaggedUnionsImpl",
    "ModPlugin",
    "ModPluginImpl",
    "ModPluginImplWithProps",
    "ModPluginWithBook",
    "PluginManager",
    "PluginNotFoundError",
    "UpdateContextImpl",
    "UpdateTemplateArgsImpl",
    "ValidateFormatTreeImpl",
    "VersionedModPlugin",
    "hookimpl",
]

import pluggy

from .book_plugin import BookPlugin
from .manager import (
    PluginManager,
    PluginNotFoundError,
)
from .mod_plugin import (
    DefaultRenderedTemplates,
    ModPlugin,
    ModPluginWithBook,
    VersionedModPlugin,
)
from .specs import (
    HEXDOC_PROJECT_NAME,
    BookPluginImpl,
    LoadTaggedUnionsImpl,
    ModPluginImpl,
    ModPluginImplWithProps,
    UpdateContextImpl,
    UpdateTemplateArgsImpl,
    ValidateFormatTreeImpl,
)
from .types import HookReturn

hookimpl = pluggy.HookimplMarker(HEXDOC_PROJECT_NAME)
"""Decorator for marking functions as hook implementations."""



import logging
import textwrap
from collections.abc import Set
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Any, Iterable, Iterator, TypeVar, cast

from pydantic import TypeAdapter
from yarl import URL

from hexdoc.core import ModResourceLoader, ResourceLocation
from hexdoc.core.properties import (
    PNGTextureOverride,
    TextureTextureOverride,
)
from hexdoc.graphics.render import BlockRenderer
from hexdoc.utils import PydanticURL
from hexdoc.utils.context import ContextSource

from ..tags import Tag
from .animated import AnimatedTexture, AnimationMeta
from .constants import MISSING_TEXTURE_URL
from .items import (
    ImageTexture,
    ItemTexture,
    MultiItemTexture,
    SingleItemTexture,
)
from .models import FoundNormalTexture, ModelItem
from .textures import PNGTexture

logger = logging.getLogger(__name__)

Texture = ImageTexture | ItemTexture

_T_Texture = TypeVar("_T_Texture", bound=Texture)


def validate_texture(
    value: Any,
    *,
    context: ContextSource,
    model_type: type[_T_Texture] | Any = Texture,
) -> _T_Texture:
    ta = TypeAdapter(model_type)
    return ta.validate_python(
        value,
        context=cast(dict[str, Any], context),  # lie
    )


class TextureNotFoundError(FileNotFoundError):
    def __init__(self, id_type: str, id: ResourceLocation):
        self.message = f"No texture found for {id_type} id: {id}"
        super().__init__(self.message)


@dataclass(kw_only=True)
class HexdocAssetLoader:
    loader: ModResourceLoader
    site_url: PydanticURL
    asset_url: PydanticURL
    render_dir: Path

    @cached_property
    def gaslighting_items(self):
        return Tag.GASLIGHTING_ITEMS.load(self.loader).value_ids_set

    @property
    def texture_props(self):
        return self.loader.props.textures

    def can_be_missing(self, id: ResourceLocation):
        if self.texture_props.missing == "*":
            return True
        return any(id.match(pattern) for pattern in self.texture_props.missing)

    def get_override(
        self,
        id: ResourceLocation,
        image_textures: dict[ResourceLocation, ImageTexture],
    ) -> Texture | None:
        match self.texture_props.override.get(id):
            case PNGTextureOverride(url=url, pixelated=pixelated):
                return PNGTexture(url=url, pixelated=pixelated)
            case TextureTextureOverride(texture=texture):
                return image_textures[texture]
            case None:
                return None

    def find_image_textures(
        self,
    ) -> Iterable[tuple[ResourceLocation, Path | ImageTexture]]:
        for resource_dir, texture_id, path in self.loader.find_resources(
            "assets",
            namespace="*",
            folder="textures",
            glob="**/*.png",
            internal_only=True,
            allow_missing=True,
        ):
            if resource_dir:
                self.loader.export_raw(
                    path=path.relative_to(resource_dir.path),
                    data=path.read_bytes(),
                )
            yield texture_id, path

    def load_item_models(self) -> Iterable[tuple[ResourceLocation, ModelItem]]:
        for _, item_id, data in self.loader.load_resources(
            "assets",
            namespace="*",
            folder="models/item",
            internal_only=True,
            allow_missing=True,
        ):
            model = ModelItem.load_data("item" / item_id, data)
            yield item_id, model

    @cached_property
    def renderer(self):
        return BlockRenderer(
            loader=self.loader,
            output_dir=self.render_dir,
        )

    def fallback_texture(self, item_id: ResourceLocation) -> ItemTexture | None:
        return None

    def load_and_render_internal_textures(
        self,
        image_textures: dict[ResourceLocation, ImageTexture],
    ) -> Iterator[tuple[ResourceLocation, Texture]]:
        """For all item/block models in all internal resource dirs, yields the item id
        (eg. `hexcasting:focus`) and some kind of texture that we can use in the book."""

        # images
        for texture_id, value in self.find_image_textures():
            if not texture_id.path.startswith("textures"):
                texture_id = "textures" / texture_id

            match value:
                case Path() as path:
                    texture = load_texture(
                        texture_id,
                        path=path,
                        repo_root=self.loader.props.repo_root,
                        asset_url=self.asset_url,
                        strict=self.texture_props.strict,
                    )

                case PNGTexture() | AnimatedTexture() as texture:
                    pass

            image_textures[texture_id] = texture
            yield texture_id, texture

        found_items_from_models = set[ResourceLocation]()
        missing_items = set[ResourceLocation]()

        missing_item_texture = SingleItemTexture.from_url(
            MISSING_TEXTURE_URL, pixelated=True
        )

        # items
        for item_id, model in self.load_item_models():
            if result := self.get_override(item_id, image_textures):
                yield item_id, result
            elif result := load_and_render_item(
                model,
                self.loader,
                self.renderer,
                self.gaslighting_items,
                image_textures,
                self.site_url,
            ):
                found_items_from_models.add(item_id)
                yield item_id, result
            else:
                missing_items.add(item_id)

        for item_id in list(missing_items):
            if result := self.fallback_texture(item_id):
                logger.warning(f"Using fallback texture for item: {item_id}")
            elif self.can_be_missing(item_id):
                logger.warning(f"Using missing texture for item: {item_id}")
                result = missing_item_texture
            else:
                continue
            missing_items.remove(item_id)
            yield item_id, result

        # oopsies
        if missing_items:
            raise FileNotFoundError(
                "Failed to find a texture for some items: "
                + ", ".join(sorted(str(item) for item in missing_items))
            )


def load_texture(
    id: ResourceLocation,
    *,
    path: Path,
    repo_root: Path,
    asset_url: URL,
    strict: bool,
) -> ImageTexture:
    # FIXME: is_relative_to is only false when reading zip archives. ideally we would
    # permalink to the gh-pages branch and copy all textures there, but we can't get
    # that commit until we build the book, so it's a bit of a circular dependency.
    if path.is_relative_to(repo_root):
        url = asset_url.joinpath(*path.relative_to(repo_root).parts)
    else:
        level = logging.WARNING if strict else logging.DEBUG
        logger.log(level, f"Failed to find relative path for {id}: {path}")
        url = None

    meta_path = path.with_suffix(".png.mcmeta")
    if meta_path.is_file():
        try:
            meta = AnimationMeta.model_validate_json(meta_path.read_bytes())
        except ValueError as e:
            logger.debug(f"Failed to parse AnimationMeta for {id}\n{e}")
        else:
            return AnimatedTexture(
                url=url,
                pixelated=True,
                css_class=id.css_class,
                meta=meta,
            )

    return PNGTexture(url=url, pixelated=True)


def load_and_render_item(
    model: ModelItem,
    loader: ModResourceLoader,
    renderer: BlockRenderer,
    gaslighting_items: Set[ResourceLocation],
    image_textures: dict[ResourceLocation, ImageTexture],
    site_url: URL,
) -> ItemTexture | None:
    try:
        match model.find_texture(loader, gaslighting_items):
            case None:
                return None

            case "gaslighting", found_textures:
                textures = list(
                    lookup_or_render_single_item(
                        found_texture,
                        renderer,
                        image_textures,
                        site_url,
                    ).inner
                    for found_texture in found_textures
                )
                return MultiItemTexture(inner=textures, gaslighting=True)

            case found_texture:
                texture = lookup_or_render_single_item(
                    found_texture,
                    renderer,
                    image_textures,
                    site_url,
                )
                return texture
    except TextureNotFoundError as e:
        logger.warning(e.message)
        return None


# TODO: move to methods on a class returned by find_texture?
def lookup_or_render_single_item(
    found_texture: FoundNormalTexture,
    renderer: BlockRenderer,
    image_textures: dict[ResourceLocation, ImageTexture],
    site_url: URL,
) -> SingleItemTexture:
    match found_texture:
        case "texture", texture_id:
            if texture_id not in image_textures:
                raise TextureNotFoundError("item", texture_id)
            return SingleItemTexture(inner=image_textures[texture_id])

        case "block_model", model_id:
            return render_block(model_id, renderer, site_url)


def render_block(
    id: ResourceLocation,
    renderer: BlockRenderer,
    site_url: URL,
) -> SingleItemTexture:
    # FIXME: hack
    id_out_path = id.path
    if id.path.startswith("item/"):
        id_out_path = "block/" + id.path.removeprefix("item/")
    elif not id.path.startswith("block/"):
        id = "block" / id

    out_path = f"assets/{id.namespace}/textures/{id_out_path}.png"

    try:
        renderer.render_block_model(id, out_path)
    except Exception as e:
        if renderer.loader.props.textures.strict:
            raise
        message = textwrap.indent(f"{e.__class__.__name__}: {e}", "  ")
        logger.error(f"Failed to render block {id}:\n{message}")
        raise TextureNotFoundError("block", id)

    logger.debug(f"Rendered {id} to {out_path}")

    # TODO: ideally we shouldn't be using site_url here, in case the site is moved
    # but I'm not sure what else we could do...
    return SingleItemTexture.from_url(site_url / out_path, pixelated=False)



import logging
from typing import Any, Iterable, Iterator

from pydantic import Field, model_validator

from hexdoc.core import ItemStack, ResourceLocation
from hexdoc.minecraft import LocalizedStr
from hexdoc.minecraft.assets import ItemWithTexture, NamedTexture
from hexdoc.model import Color, IDModel
from hexdoc.patchouli.page.abstract_pages import AccumulatorPage, PageWithAccumulator
from hexdoc.utils import Sortable

from .page import Page
from .utils import AdvancementSpoilered, Flagged

logger = logging.getLogger(__name__)


class Entry(IDModel, Sortable, AdvancementSpoilered, Flagged):
    """Entry json file, with pages and localizations.

    See: https://vazkiimods.github.io/Patchouli/docs/reference/entry-json
    """

    # required (entry.json)
    name: LocalizedStr
    category_id: ResourceLocation = Field(alias="category")
    icon: ItemWithTexture | NamedTexture
    pages: list[Page]

    # optional (entry.json)
    advancement: ResourceLocation | None = None
    priority: bool = False
    secret: bool = False
    read_by_default: bool = False
    sortnum: int = 0
    turnin: ResourceLocation | None = None
    extra_recipe_mappings: dict[ItemStack, int] | None = None
    entry_color: Color | None = None  # this is undocumented lmao

    @property
    def _cmp_key(self) -> tuple[bool, int, LocalizedStr]:
        # implement Sortable
        # note: python sorts false before true, so we invert priority
        return (not self.priority, self.sortnum, self.name)

    @property
    def anchors(self) -> Iterable[str]:
        for page in self.pages:
            if page.anchor is not None:
                yield page.anchor

    @property
    def book_link_key(self):
        """Key to look up this entry in `BookContext.book_links`."""
        return str(self.id)

    @property
    def fragment(self):
        """URL fragment for this entry in `BookContext.book_links`."""
        return self.id.path

    @property
    def redirect_path(self):
        """Path to this entry when generating redirect pages."""
        return self.id.path

    @property
    def first_text_page(self):
        for page in self.pages:
            if getattr(page, "text", None):
                return page

    def preprocess_pages(self) -> Iterator[Page]:
        """Combines adjacent PageWithAccumulator recipes as much as possible."""
        acc: AccumulatorPage[Any] | None = None

        for page in self.pages:
            if isinstance(page, PageWithAccumulator):
                if not (acc and acc.can_append(page)):
                    if acc and acc.has_content:
                        yield acc
                    acc = page.accumulator_type().from_page(page)

                acc.append(page)

                if not acc.can_append_more:
                    yield acc
                    acc = None
            else:
                if acc and acc.has_content:
                    yield acc
                acc = None
                yield page

        if acc and acc.has_content:
            yield acc

    def _get_advancement(self):
        # implements AdvancementSpoilered
        return self.advancement

    @model_validator(mode="after")
    def _skip_disabled_pages(self):
        new_pages = list[Page]()
        for i, page in enumerate(self.pages):
            if not page.is_flag_enabled:
                logger.info(
                    f"Skipping page {i} of entry {self.id} due to disabled flag {page.flag}"
                )
                continue
            new_pages.append(page)
        if not new_pages:
            logger.warning(
                f"Entry has no pages{' after applying flags' if self.pages else ''}: {self.id}"
            )
        self.pages = new_pages
        return self



import subprocess
import textwrap
from pathlib import Path
from subprocess import CalledProcessError
from typing import Callable, Sequence, TypeVar

_T = TypeVar("_T")


def run(
    args: str | Sequence[str],
    *,
    cwd: str | Path | None = None,
    env: dict[str, str] | None = None,
    type: Callable[[str], _T] = str,
) -> _T:
    """Runs a command using `subprocess.run`, returning the stripped stdout as UTF-8.

    If `type` is provided, converts the stripped stdout to that type.

    If the command fails, raises `CalledProcessError` with stdout/stderr as notes.
    """
    try:
        result = subprocess.run(
            args=args,
            cwd=cwd,
            env=env,
            check=True,
            capture_output=True,
            encoding="utf-8",
        )
        return type(result.stdout.strip())
    except CalledProcessError as e:
        for name, value in [("stdout", e.stdout), ("stderr", e.stderr)]:
            if value:
                e.add_note(f"{name}:\n" + textwrap.indent(str(value).strip(), "  "))
        raise



from pydantic import Field
from yarl import URL

from hexdoc.core import ResourceLocation
from hexdoc.core.resource_dir import PathResourceDir
from hexdoc.data import HexdocMetadata
from hexdoc.model import ValidationContextModel
from hexdoc.patchouli.text import BookLinks


class BookContext(ValidationContextModel):
    modid: str
    book_id: ResourceLocation
    book_links: BookLinks = Field(default_factory=dict)
    spoilered_advancements: set[ResourceLocation]
    all_metadata: dict[str, HexdocMetadata]
    flags: dict[str, bool]

    def get_link_base(self, resource_dir: PathResourceDir) -> URL:
        modid = resource_dir.modid
        if resource_dir.internal or modid is None or modid == self.modid:
            return URL()

        book_url = self.all_metadata[modid].book_url
        if book_url is None:
            raise ValueError(f"Mod {modid} does not export a book url")

        return book_url
